from multiprocessing import Pool
from typing import Any, Callable, Dict, List, TypeVar

import cloudpickle
from triad.utils.convert import to_timedelta
from tune.constants import TUNE_STOPPER_DEFAULT_CHECK_INTERVAL
from tune.exceptions import TuneInterrupted

T = TypeVar("T")


def run_monitored_process(
    target: Callable[..., T],
    args: List[Any],
    kwargs: Dict[str, Any],
    stop_checker: Callable[[], bool],
    interval: Any = TUNE_STOPPER_DEFAULT_CHECK_INTERVAL,
) -> T:
    interval_sec = to_timedelta(interval).total_seconds()
    blob = cloudpickle.dumps((target, args, kwargs))
    pool = Pool(1)
    result = pool.map_async(_run_target, [blob])
    while True:
        result.wait(interval_sec)
        if result.ready():
            pool.close()
            pool.join()
            res = result.get(timeout=1)[0]
            return cloudpickle.pickle.loads(res)
        if stop_checker():
            pool.terminate()
            pool.join()
            raise TuneInterrupted()


def _run_target(blob: Any):
    tp = cloudpickle.loads(blob)
    return cloudpickle.dumps(tp[0](*tp[1], **tp[2]))



from datetime import datetime
from threading import RLock
from typing import Any, Callable, Dict, List, Set

from triad.utils.convert import to_timedelta
from tune.concepts.flow import (
    Trial,
    TrialDecision,
    TrialJudge,
    TrialReport,
    TrialReportLogger,
)


class TrialReportCollection(TrialReportLogger):
    def __init__(self, new_best_only: bool = False):
        super().__init__(new_best_only=new_best_only)
        self._reports: List[TrialReport] = []

    def log(self, report: TrialReport) -> None:
        self._reports.append(report.reset_log_time())

    @property
    def reports(self) -> List[TrialReport]:
        with self._lock:
            return list(self._reports)


class NonIterativeStopper(TrialJudge):
    def __init__(self, log_best_only: bool = False):
        super().__init__()
        self._stopper_updated = False
        self._lock = RLock()
        self._log_best_only = log_best_only
        self._logs: Dict[str, TrialReportCollection] = {}

    @property
    def updated(self) -> bool:
        return self._stopper_updated

    def should_stop(self, trial: Trial) -> bool:  # pragma: no cover
        return False

    def on_report(self, report: TrialReport) -> bool:
        self._stopper_updated = True
        self.monitor.on_report(report)
        with self._lock:
            key = str(report.trial.keys)
            if key not in self._logs:
                self._logs[key] = TrialReportCollection(self._log_best_only)
            return self._logs[key].on_report(report)

    def can_accept(self, trial: Trial) -> bool:
        return not self.should_stop(trial)

    def judge(self, report: TrialReport) -> TrialDecision:
        self.on_report(report)
        return TrialDecision(report, 0.0, False)

    def get_reports(self, trial: Trial) -> List[TrialReport]:
        with self._lock:
            key = str(trial.keys)
            if key not in self._logs:
                return []
            v = self._logs[key]
        return v.reports

    def __and__(self, other: "NonIterativeStopper") -> "NonIterativeStopperCombiner":
        return NonIterativeStopperCombiner(self, other, is_and=True)

    def __or__(self, other: "NonIterativeStopper") -> "NonIterativeStopperCombiner":
        return NonIterativeStopperCombiner(self, other, is_and=False)


class NonIterativeStopperCombiner(NonIterativeStopper):
    def __init__(
        self, left: NonIterativeStopper, right: NonIterativeStopper, is_and: bool
    ):
        super().__init__()
        assert not left.updated, "can't reuse updated stopper"
        assert not right.updated, "can't reuse updated stopper"
        self._left = left
        self._right = right
        self._is_and = is_and

    def should_stop(self, trial: Trial) -> bool:  # pragma: no cover
        if self._is_and:
            return self._left.should_stop(trial) and self._right.should_stop(trial)
        else:
            return self._left.should_stop(trial) or self._right.should_stop(trial)

    def on_report(self, report: TrialReport) -> bool:
        self.monitor.on_report(report)
        left = self._left.on_report(report)
        right = self._right.on_report(report)
        return left or right

    def get_reports(self, trial: Trial) -> List[TrialReport]:  # pragma: no cover
        raise NotImplementedError


class SimpleNonIterativeStopper(NonIterativeStopper):
    def __init__(
        self,
        partition_should_stop: Callable[[TrialReport, bool, List[TrialReport]], bool],
        log_best_only: bool = False,
    ):
        super().__init__(log_best_only=log_best_only)
        self._partition_should_stop = partition_should_stop
        self._stopped: Set[str] = set()

    def should_stop(self, trial: Trial) -> bool:
        key = str(trial.keys)
        with self._lock:
            return key in self._stopped

    def on_report(self, report: TrialReport) -> bool:
        updated = super().on_report(report)
        key = str(report.trial.keys)
        with self._lock:
            if key not in self._stopped:
                if self._partition_should_stop(
                    report, updated, self.get_reports(report.trial)
                ):
                    self._stopped.add(key)
        return updated


def n_samples(n: int) -> SimpleNonIterativeStopper:
    def func(current: TrialReport, updated: bool, reports: List[TrialReport]):
        return len(reports) >= n

    return SimpleNonIterativeStopper(func, log_best_only=False)


def n_updates(n: int) -> SimpleNonIterativeStopper:
    def func(current: TrialReport, updated: bool, reports: List[TrialReport]):
        return len(reports) >= n

    return SimpleNonIterativeStopper(func, log_best_only=True)


def no_update_period(period: Any) -> SimpleNonIterativeStopper:
    _interval = to_timedelta(period)

    def func(current: TrialReport, updated: bool, reports: List[TrialReport]):
        if updated or len(reports) == 0:
            return False
        return datetime.now() - reports[-1].log_time > _interval

    return SimpleNonIterativeStopper(func, log_best_only=True)


def small_improvement(threshold: float, updates: int) -> SimpleNonIterativeStopper:
    assert updates > 0

    def func(current: TrialReport, updated: bool, reports: List[TrialReport]):
        if not updated:
            return False
        if len(reports) <= updates:
            return False
        diff = reports[-updates - 1].sort_metric - current.sort_metric
        return diff < threshold

    return SimpleNonIterativeStopper(func, log_best_only=True)



from fs.base import FS as FSBase
from triad import FileSystem
from tune.iterative.objective import (
    IterativeObjectiveFunc,
    validate_iterative_objective,
)
from tune.concepts.flow import Trial, TrialDecision, TrialJudge, TrialReport, Monitor


class F(IterativeObjectiveFunc):
    def __init__(self):
        self.v = -10
        super().__init__()

    def copy(self) -> "F":
        return F()

    def initialize(self) -> None:
        self.v = 0

    def finalize(self) -> None:
        self.v = -10

    def load_checkpoint(self, fs: FSBase) -> None:
        self.v = int(fs.readtext("x"))

    def save_checkpoint(self, fs: FSBase) -> None:
        fs.writetext("x", str(self.v))

    def run_single_iteration(self) -> TrialReport:
        self.v += 1
        return TrialReport(self.current_trial, self.v, metadata={"d": 4})

    def generate_sort_metric(self, value: float) -> float:
        return -value


class J(TrialJudge):
    def __init__(self, schedule):
        super().__init__()
        self.schedule = schedule

    def can_accept(self, trial: Trial) -> bool:
        return True

    def get_budget(self, trial: Trial, rung: int) -> float:
        return float(self.schedule[rung]) if rung < len(self.schedule) else 0.0

    def judge(self, report: TrialReport) -> TrialDecision:
        self.report = report
        return TrialDecision(
            report,
            budget=0,  # stop at each rung
            should_checkpoint=True,
            metadata={"x": 1},
        )


class M(Monitor):
    def __init__(self):
        self._reports = []

    def on_report(self, report: TrialReport) -> None:
        return self._reports.append(report)


def test_objective_func(tmpdir):
    fs = FileSystem().opendir(str(tmpdir))
    j = J([3, 3, 2])
    f = F().copy()
    t = Trial("abc", {"a": 1})
    f.run(t, judge=j, checkpoint_basedir_fs=fs)
    assert -10 == f.v
    f.run(t, judge=j, checkpoint_basedir_fs=fs)
    assert -10 == f.v
    assert 6.0 == j.report.metric
    assert -6.0 == j.report.sort_metric
    f.run(t, judge=j, checkpoint_basedir_fs=fs)
    assert -10 == f.v
    assert 8.0 == j.report.metric
    assert -8.0 == j.report.sort_metric


def test_validator():
    m = M()
    for cont in [True, False]:
        validate_iterative_objective(
            F(),
            Trial("abc", {"a": 1}),
            [3, 3, 2],
            lambda reports: [-3.0, -6.0, -8.0] == [x.sort_metric for x in reports],
            continuous=cont,
            monitor=m,
        )
    assert 6 == len(m._reports)



import os
import tempfile
from typing import Callable, List, Optional
from uuid import uuid4

from cloudpickle import pickle
from fs.base import FS as FSBase
from triad import FileSystem
from tune.concepts.checkpoint import Checkpoint
from tune.concepts.flow import Monitor, Trial, TrialDecision, TrialJudge, TrialReport


class IterativeObjectiveFunc:
    def __init__(self):
        self._rung = 0
        self._current_trial: Optional[Trial] = None

    def copy(self) -> "IterativeObjectiveFunc":  # pragma: no cover
        raise NotImplementedError

    @property
    def current_trial(self) -> Trial:
        assert self._current_trial is not None
        return self._current_trial

    @property
    def rung(self) -> int:
        return self._rung

    def generate_sort_metric(self, value: float) -> float:
        return value

    def load_checkpoint(self, fs: FSBase) -> None:  # pragma: no cover
        return

    def save_checkpoint(self, fs: FSBase) -> None:  # pragma: no cover
        return

    def initialize(self) -> None:  # pragma: no cover
        return

    def finalize(self) -> None:  # pragma: no cover
        return

    def run_single_iteration(self) -> TrialReport:  # pragma: no cover
        raise NotImplementedError

    def run_single_rung(self, budget: float) -> TrialReport:
        used = 0.0
        while True:
            current_report = self.run_single_iteration()
            used += current_report.cost
            if used >= budget:
                return current_report.with_cost(used)

    def run(
        self,
        trial: Trial,
        judge: TrialJudge,
        checkpoint_basedir_fs: FSBase,
    ) -> None:
        checkpoint = Checkpoint(
            checkpoint_basedir_fs.makedir(trial.trial_id, recreate=True)
        )
        if not judge.can_accept(trial):
            return
        self._current_trial = trial
        self.initialize()
        try:
            if len(checkpoint) > 0:
                self._rung = int(checkpoint.latest.readtext("__RUNG__")) + 1
                self.load_checkpoint(checkpoint.latest)
            budget = judge.get_budget(trial, self.rung)
            while budget > 0:
                report = self.run_single_rung(budget)
                report = report.with_rung(self.rung).with_sort_metric(
                    self.generate_sort_metric(report.metric)
                )
                decision = judge.judge(report)
                if decision.should_checkpoint:
                    with checkpoint.create() as fs:
                        fs.writetext("__RUNG__", str(self.rung))
                        self.save_checkpoint(fs)
                budget = decision.budget
                self._rung += 1
        finally:
            self.finalize()


def validate_iterative_objective(
    func: IterativeObjectiveFunc,
    trial: Trial,
    budgets: List[float],
    validator: Callable[[List[TrialReport]], None],
    continuous: bool = False,
    checkpoint_path: str = "",
    monitor: Optional[Monitor] = None,
) -> None:
    path = checkpoint_path if checkpoint_path != "" else tempfile.gettempdir()
    basefs = FileSystem().makedirs(os.path.join(path, str(uuid4())), recreate=True)
    j = _Validator(monitor, budgets, continuous=continuous)
    if continuous:
        f = pickle.loads(pickle.dumps(func)).copy()
        f.run(trial, j, checkpoint_basedir_fs=basefs)
    else:
        for _ in budgets:
            f = pickle.loads(pickle.dumps(func)).copy()
            f.run(trial, j, checkpoint_basedir_fs=basefs)
    validator(j.reports)


class _Validator(TrialJudge):
    def __init__(
        self, monitor: Optional[Monitor], budgets: List[float], continuous: bool
    ):
        super().__init__(monitor)
        self._budgets = budgets
        self._continuous = continuous
        self._reports: List[TrialReport] = []

    @property
    def reports(self) -> List[TrialReport]:
        return self._reports

    def can_accept(self, trial: Trial) -> bool:

        return True

    def get_budget(self, trial: Trial, rung: int) -> float:
        budget = self._budgets[rung] if rung < len(self._budgets) else 0.0
        self.monitor.on_get_budget(trial, rung, budget)
        return budget

    def judge(self, report: TrialReport) -> TrialDecision:
        self.monitor.on_report(report)
        self._reports.append(report)
        decision = TrialDecision(
            report,
            budget=self.get_budget(report.trial, report.rung + 1)
            if self._continuous
            else 0.0,
            should_checkpoint=report.rung >= len(self._budgets)
            if self._continuous
            else True,
        )
        self.monitor.on_judge(decision)
        return decision



from tune import Checkpoint
from triad import FileSystem
from pytest import raises


def test_checkpoint(tmpdir):
    fs = FileSystem().opendir(str(tmpdir))
    cp = Checkpoint(fs)
    assert 0 == len(cp)
    with raises(AssertionError):
        cp.latest
    try:
        for i in range(4):
            with cp.create() as sfs:
                sfs.writetext("a.txt", str(i))
                if i == 3:
                    raise Exception
    except Exception:
        pass
    assert 3 == len(cp)
    assert "2" == cp.latest.readtext("a.txt")
    files = fs.listdir(".")
    assert 4 == len(files)
    cp2 = Checkpoint(fs)
    assert 3 == len(cp2)
    assert "2" == cp2.latest.readtext("a.txt")



from typing import Optional

from triad import FileSystem
from tune.iterative.objective import IterativeObjectiveFunc
from tune.noniterative.objective import NonIterativeObjectiveFunc
from tune.concepts.flow import Trial, TrialDecision, TrialJudge, TrialReport


class _NonIterativeObjectiveWrapper(NonIterativeObjectiveFunc):
    def __init__(
        self,
        func: IterativeObjectiveFunc,
        checkpoint_path: str,
        budget: float,
    ):
        super().__init__()
        self._budget = budget
        self._func = func
        self._checkpoint_path = checkpoint_path

    def generate_sort_metric(self, value: float) -> float:
        return self._func.generate_sort_metric(value)

    def run(self, trial: Trial) -> TrialReport:  # pragma: no cover
        judge = _NonIterativeJudgeWrapper(self._budget)
        fs = FileSystem().makedirs(self._checkpoint_path, recreate=True)
        self._func = self._func.copy()
        self._func.run(trial, judge=judge, checkpoint_basedir_fs=fs)
        return judge.report


class _NonIterativeJudgeWrapper(TrialJudge):
    def __init__(self, budget: float):
        super().__init__()
        self._report: Optional[TrialReport] = None
        self._budget = budget

    @property
    def report(self) -> TrialReport:
        assert self._report is not None
        return self._report

    def can_accept(self, trial: Trial) -> bool:
        return True

    def get_budget(self, trial: Trial, rung: int):
        return self._budget

    def judge(self, report: TrialReport) -> TrialDecision:
        self._report = report
        return TrialDecision(report, budget=0.0, should_checkpoint=True, metadata={})



import json
import os
import pickle
import random
from typing import Any, Dict, Iterable, List, Optional, Tuple
from uuid import uuid4

import numpy as np
import pandas as pd
from fugue import (
    ArrayDataFrame,
    DataFrame,
    ExecutionEngine,
    FugueWorkflow,
    IterableDataFrame,
    LocalDataFrame,
    Transformer,
    WorkflowDataFrame,
    WorkflowDataFrames,
)
from triad import ParamDict, assert_or_throw, to_uuid

from tune.constants import (
    TUNE_DATASET_DF_PREFIX,
    TUNE_DATASET_PARAMS_PREFIX,
    TUNE_DATASET_TRIALS,
    TUNE_PREFIX,
    TUNE_REPORT,
    TUNE_REPORT_ID,
    TUNE_REPORT_METRIC,
    TUNE_TEMP_PATH,
)
from tune.exceptions import TuneCompileError
from tune.concepts.space import Space
from tune.concepts.flow import Trial


class TuneDataset:
    def __init__(self, data: WorkflowDataFrame, dfs: List[str], keys: List[str]):
        self._data = data.persist()
        self._dfs = dfs
        self._keys = keys

    @property
    def data(self) -> WorkflowDataFrame:
        return self._data

    @property
    def dfs(self) -> List[str]:
        return self._dfs

    @property
    def keys(self) -> List[str]:
        return self._keys

    def divide(self, weights: List[float], seed: Any) -> List["TuneDataset"]:
        def label(df: pd.DataFrame) -> pd.DataFrame:
            if seed is not None:
                np.random.seed(seed)
            w = np.array(weights)
            p = w / np.sum(w)
            df["__tune_divide_id_"] = np.random.choice(len(weights), df.shape[0], p=p)
            return df.reset_index(drop=True)

        def select(df: pd.DataFrame, n: int) -> pd.DataFrame:
            return (
                df[df["__tune_divide_id_"] == n]
                .drop(["__tune_divide_id_"], axis=1)
                .reset_index(drop=True)
            )

        temp = self._data.process(label).persist()
        datasets: List["TuneDataset"] = []
        for i in range(len(weights)):
            datasets.append(
                TuneDataset(
                    temp.transform(
                        select, schema="*-__tune_divide_id_", params=dict(n=i)
                    ),
                    self.dfs,
                    self.keys,
                )
            )
        return datasets


def get_trials_from_row(row: Dict[str, Any], with_dfs: bool = True) -> Iterable[Trial]:
    if not with_dfs:
        for params in json.loads(row[TUNE_DATASET_TRIALS]):
            yield Trial.from_jsondict(params)
    else:
        dfs: Dict[str, Any] = {}
        for k, v in row.items():
            if k.startswith(TUNE_DATASET_DF_PREFIX):
                key = k[len(TUNE_DATASET_DF_PREFIX) :]
                if v is not None:
                    dfs[key] = pd.read_parquet(v)
        for params in json.loads(row[TUNE_DATASET_TRIALS]):
            yield Trial.from_jsondict(params).with_dfs(dfs)


class TuneDatasetBuilder:
    def __init__(self, space: Space, path: str = ""):
        self._dfs_spec: List[Tuple[str, WorkflowDataFrame, str]] = []
        self._space = space
        self._path = path

    def add_df(
        self, name: str, df: WorkflowDataFrame, how: str = ""
    ) -> "TuneDatasetBuilder":
        assert_or_throw(
            not any(r[0] == name for r in self._dfs_spec),
            TuneCompileError(name + " already exists"),
        )
        if len(self._dfs_spec) == 0:
            assert_or_throw(
                how == "", TuneCompileError("first dataframe can't specify how to join")
            )
        else:
            assert_or_throw(
                how != "",
                TuneCompileError("must specify how to join after first dataframe"),
            )
        self._dfs_spec.append((name, df, how))
        return self

    def add_dfs(self, dfs: WorkflowDataFrames, how: str = "") -> "TuneDatasetBuilder":
        assert_or_throw(dfs.has_key, "all datarames must be named")
        for k, v in dfs.items():
            if len(self._dfs_spec) == 0:
                self.add_df(k, v)
            else:
                self.add_df(k, v, how=how)
        return self

    def build(
        self,
        wf: FugueWorkflow,
        batch_size: int = 1,
        shuffle: bool = True,
        trial_metadata: Optional[Dict[str, Any]] = None,
    ) -> TuneDataset:
        space = self._space_to_df(wf=wf, batch_size=batch_size, shuffle=shuffle)
        if len(self._dfs_spec) == 0:
            res = space
            keys: List[str] = []
        else:
            dfs, keys = self._serialize_dfs()
            res = dfs.cross_join(space)

        def finalize(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
            for row in df:
                yield _to_trail_row(row, trial_metadata or {})

        data = res.transform(
            finalize,
            schema=f"*,{TUNE_DATASET_TRIALS}:str-{TUNE_DATASET_PARAMS_PREFIX}",
        )
        return TuneDataset(data, [x[0] for x in self._dfs_spec], keys)

    def _serialize_df(self, df: WorkflowDataFrame, name: str) -> WorkflowDataFrame:
        pre_partition = df.partition_spec
        path = self._path

        def _get_temp_path(p: str, conf: ParamDict) -> str:
            if p is not None and p != "":
                return p
            return conf.get_or_throw(TUNE_TEMP_PATH, str)

        if len(pre_partition.partition_by) == 0:

            def save_single_file(e: ExecutionEngine, _input: DataFrame) -> DataFrame:
                p = _get_temp_path(path, e.conf)
                fp = os.path.join(p, str(uuid4()) + ".parquet")
                e.save_df(_input, fp, force_single=True)
                return ArrayDataFrame([[fp]], f"{TUNE_DATASET_DF_PREFIX}{name}:str")

            return df.process(save_single_file)
        else:

            class SavePartition(Transformer):
                def get_output_schema(self, df: DataFrame) -> Any:
                    dfn = self.params.get_or_throw("name", str)
                    return self.key_schema + f"{TUNE_DATASET_DF_PREFIX}{dfn}:str"

                def transform(self, df: LocalDataFrame) -> LocalDataFrame:
                    p = _get_temp_path(self.params.get("path", ""), self.workflow_conf)
                    fp = os.path.join(p, str(uuid4()) + ".parquet")
                    first = df.peek_dict()
                    keys = [first[x] for x in self.key_schema.names]
                    df.as_pandas().to_parquet(fp)
                    return ArrayDataFrame([keys + [fp]], self.output_schema)

            return df.transform(SavePartition, params={"path": path, "name": name})

    def _serialize_dfs(self) -> Tuple[WorkflowDataFrame, List[str]]:
        df = self._serialize_df(self._dfs_spec[0][1], self._dfs_spec[0][0])
        keys = list(self._dfs_spec[0][1].partition_spec.partition_by)
        for i in range(1, len(self._dfs_spec)):
            how = self._dfs_spec[i][2]
            new_keys = set(self._dfs_spec[i][1].partition_spec.partition_by)
            next_df = self._serialize_df(self._dfs_spec[i][1], self._dfs_spec[i][0])
            df = df.join(next_df, how=how)
            if how != "cross":
                keys = [k for k in keys if k in new_keys]
        return df, keys

    def _space_to_df(
        self, wf: FugueWorkflow, batch_size: int = 1, shuffle: bool = True
    ) -> WorkflowDataFrame:
        def get_data() -> Iterable[List[Any]]:
            it = list(self._space)  # type: ignore
            if shuffle:
                random.seed(0)
                random.shuffle(it)
            res: List[Any] = []
            for a in it:
                res.append(a)
                if batch_size == len(res):
                    yield [pickle.dumps(res)]
                    res = []
            if len(res) > 0:
                yield [pickle.dumps(res)]

        return wf.df(
            IterableDataFrame(get_data(), f"{TUNE_DATASET_PARAMS_PREFIX}:binary")
        )


class StudyResult:
    def __init__(self, dataset: TuneDataset, result: WorkflowDataFrame):
        self._dataset = dataset
        self._result = (
            result.persist()
            .partition_by(TUNE_REPORT_ID, presort=TUNE_REPORT_METRIC)
            .take(1)
            .persist()
        )

    def result(self, best_n: int = 0) -> WorkflowDataFrame:
        if best_n <= 0:
            return self._result
        if len(self._dataset.keys) == 0:
            return self._result.take(n=best_n, presort=TUNE_REPORT_METRIC)
        else:
            return self._result.partition(
                by=self._dataset.keys, presort=TUNE_REPORT_METRIC
            ).take(best_n)

    def next_tune_dataset(self, best_n: int = 0) -> TuneDataset:
        data = self.result(best_n).drop(
            [TUNE_REPORT_ID, TUNE_REPORT_METRIC, TUNE_REPORT], if_exists=True
        )
        return TuneDataset(data, dfs=self._dataset.dfs, keys=self._dataset.keys)

    def union_with(self, other: "StudyResult") -> None:
        self._result = (
            self._result.union(other._result)
            .partition_by(TUNE_REPORT_ID, presort=TUNE_REPORT_METRIC)
            .take(1)
            .persist()
        )


def _to_trail_row(data: Dict[str, Any], metadata: Dict[str, Any]) -> Dict[str, Any]:
    key_names = sorted(k for k in data.keys() if not k.startswith(TUNE_PREFIX))
    keys = [data[k] for k in key_names]
    trials: Dict[str, Dict[str, Any]] = {}
    for param in pickle.loads(data[TUNE_DATASET_PARAMS_PREFIX]):
        p = ParamDict(sorted(((k, v) for k, v in param.items()), key=lambda x: x[0]))
        tid = to_uuid(keys, p)
        trials[tid] = Trial(
            trial_id=tid, params=p, metadata=metadata, keys=keys
        ).jsondict
    data[TUNE_DATASET_TRIALS] = json.dumps(list(trials.values()))
    del data[TUNE_DATASET_PARAMS_PREFIX]
    return data



from threading import RLock
from typing import Any, Callable, Dict, List, Tuple
from unittest import TestCase

from tune import (
    Choice,
    NonIterativeObjectiveLocalOptimizer,
    Rand,
    RandInt,
    StochasticExpression,
    Trial,
    noniterative_objective,
)
from tune._utils import assert_close
from tune.noniterative.objective import validate_noniterative_objective


class NonIterativeObjectiveLocalOptimizerTests(object):
    """DataFrame level general test suite.
    All new DataFrame types should pass this test suite.
    """

    class Tests(TestCase):
        def make_optimizer(self, **kwargs: Any) -> NonIterativeObjectiveLocalOptimizer:
            raise NotImplementedError

        def test_choice(self):
            d = dict(a=1, b=1, c=1)
            values = self._generate_values(Choice("a", "b", "c"), lambda x: d[x])
            assert len(values) > 0
            assert all(x in ["a", "b", "c"] for x in values)
            assert all(c in values for c in ["a", "b", "c"])

        def test_rand(self):
            # common case
            values = self._generate_values(Rand(-2.0, 3.0), lambda x: x ** 2)
            assert len(values) > 0
            assert all(x >= -2.0 and x <= 3.0 for x in values)

            # with q, and range%q == 0
            values = self._generate_values(Rand(-2.0, 3.0, q=2.5), lambda x: x ** 2)
            assert_close(values, [-2.0, 0.5, 3.0])

            values = self._generate_values(
                Rand(-2.0, 3.0, q=2.5, include_high=False), lambda x: x ** 2
            )
            assert_close(values, [-2.0, 0.5])

            # with q, and range%q != 0
            for ih in [True, False]:
                values = self._generate_values(
                    Rand(-2.0, 3.0, q=3.0, include_high=ih), lambda x: x ** 2
                )
                assert_close(values, [-2.0, 1.0])

            # with log
            values = self._generate_values(Rand(1.0, 3.0, log=True), lambda x: x ** 2)
            assert all(x >= 1.0 and x <= 3.0 for x in values)

            # with log and q, and range%q == 0
            values = self._generate_values(
                Rand(1.0, 6.0, q=2.5, log=True), lambda x: x ** 2
            )
            assert_close(values, [1.0, 3.5, 6.0])

            values = self._generate_values(
                Rand(1.0, 6.0, q=2.5, log=True, include_high=False), lambda x: x ** 2
            )
            assert_close(values, [1.0, 3.5])

            # with log and q, and range%q != 0
            for ih in [True, False]:
                values = self._generate_values(
                    Rand(1.0, 6.0, q=3.0, log=True, include_high=ih), lambda x: x ** 2
                )
                assert_close(values, [1.0, 4.0])

        def test_randint(self):
            for log in [True, False]:
                # common case
                values = self._generate_values(RandInt(1, 3, log=log), lambda x: x ** 2)
                assert_close(values, [1, 2, 3])
                assert all(isinstance(x, int) for x in values)

                values = self._generate_values(
                    RandInt(1, 3, include_high=False, log=log), lambda x: x ** 2
                )
                assert_close(values, [1, 2])

                # with q, range % q != 0
                for ih in [True, False]:
                    values = self._generate_values(
                        RandInt(1, 6, 2, include_high=ih, log=log), lambda x: x ** 2
                    )
                    assert_close(values, [1, 3, 5])
                    assert all(isinstance(x, int) for x in values)

                # with q, range % q == 0
                values = self._generate_values(
                    RandInt(1, 5, 2, log=log), lambda x: x ** 2
                )
                assert_close(values, [1, 3, 5])

                values = self._generate_values(
                    RandInt(1, 5, 2, include_high=False, log=log), lambda x: x ** 2
                )
                assert_close(values, [1, 3])

        def test_optimization_dummy(self):
            params = dict(a=1, b=2, c=3)
            trial = Trial("a", params, metadata={})
            o = self.make_optimizer(max_iter=5)

            @noniterative_objective
            def objective(a, b, c) -> Tuple[float, Dict[str, Any]]:
                return a ** 2 + b ** 2 + c, dict(a="x")

            def v(report):
                assert 1 == report.params["a"]
                assert 2 == report.params["b"]
                assert 3 == report.params["c"]
                assert report.metric == 8
                assert "x" == report.metadata["a"]

            validate_noniterative_objective(objective, trial, v, optimizer=o)

        def test_optimization(self):
            params = dict(a=Rand(-10.0, 10.0), b=RandInt(-100, 100), c=2.0)
            trial = Trial("a", params, metadata={})
            o = self.make_optimizer(max_iter=200)

            @noniterative_objective
            def objective(a, b, c) -> Tuple[float, Dict[str, Any]]:
                return a ** 2 + b ** 2 + c, dict(a="x")

            def v(report):
                print(report.metric)
                assert report.metric < 7
                assert report.params["a"] ** 2 < 2
                assert report.params["b"] ** 2 < 2
                assert 2.0 == report.params["c"]
                assert "x" == report.metadata["a"]

            validate_noniterative_objective(objective, trial, v, optimizer=o)

        def _generate_values(
            self, expr: StochasticExpression, obj: Callable[..., float]
        ) -> List[Any]:
            params = dict(a=expr)
            trial = Trial("x", params, metadata={})
            o = self.make_optimizer(max_iter=30)
            lock = RLock()
            values: List[Any] = []

            @noniterative_objective
            def objective(a: Any) -> float:
                with lock:
                    values.append(a)
                return obj(a)

            o.run(objective, trial)  # type: ignore
            return values



from typing import Dict, Optional, Type

from fs.base import FS as FSBase
from tensorflow import keras
from tune import IterativeObjectiveFunc, TrialReport

from tune_tensorflow.spec import KerasTrainingSpec
from tune_tensorflow.utils import extract_keras_spec


class KerasObjective(IterativeObjectiveFunc):
    def __init__(self, type_dict: Dict[str, Type[KerasTrainingSpec]]) -> None:
        super().__init__()
        self._epochs = 0
        self._spec: Optional[KerasTrainingSpec] = None
        self._model: Optional[keras.models.Model] = None
        self._type_dict = type_dict

    @property
    def model(self) -> keras.models.Model:
        assert self._model is not None
        return self._model

    @property
    def spec(self) -> KerasTrainingSpec:
        assert self._spec is not None
        return self._spec

    def copy(self) -> "KerasObjective":
        return KerasObjective(self._type_dict)

    def generate_sort_metric(self, value: float) -> float:
        return self.spec.generate_sort_metric(value)

    def save_checkpoint(self, fs: FSBase) -> None:
        self.spec.save_checkpoint(fs, self.model)
        fs.writetext("epoch", str(self._epochs))

    def load_checkpoint(self, fs: FSBase) -> None:
        self.spec.load_checkpoint(fs, self.model)
        self._epochs = int(fs.readtext("epoch"))

    def run_single_rung(self, budget: float) -> TrialReport:
        trial = self.current_trial
        fit_args, fit_kwargs = self.spec.get_fit_params()
        fit_kwargs = dict(fit_kwargs)
        fit_kwargs.update(
            dict(epochs=self._epochs + int(budget), initial_epoch=self._epochs)
        )
        h = self.model.fit(*fit_args, **fit_kwargs)
        metric = self.spec.get_fit_metric(h)
        self._epochs += int(budget)
        return TrialReport(trial=trial, metric=metric, cost=budget, rung=self.rung)

    def initialize(self) -> None:
        spec = extract_keras_spec(self.current_trial.params, self._type_dict)
        self._spec = spec(self.current_trial.params, self.current_trial.dfs)
        self._model = self.spec.compile_model()

    def finalize(self) -> None:
        self.spec.finalize()



# flake8: noqa

from tune.concepts.flow.judge import (
    Monitor,
    NoOpTrailJudge,
    RemoteTrialJudge,
    TrialCallback,
    TrialDecision,
    TrialJudge,
)
from tune.concepts.flow.report import TrialReport, TrialReportHeap, TrialReportLogger
from tune.concepts.flow.trial import Trial



import numpy as np
from scipy import stats
from tune._utils import (
    assert_close,
    normal_to_continuous,
    normal_to_discrete,
    normal_to_integers,
    uniform_to_choice,
    uniform_to_continuous,
    uniform_to_discrete,
    uniform_to_integers,
)


def test_uniform_to_continuous():
    assert 10 == uniform_to_continuous(0.5, 10, 10)
    assert 10 == uniform_to_continuous(0.5, 10, 9)
    assert 12.5 == uniform_to_continuous(0.5, 10, 15)
    assert 10 == uniform_to_continuous(0.0, 10, 15)
    assert 15 == uniform_to_continuous(1.0, 10, 15)  # impossible case

    np.random.seed(0)
    values = np.random.uniform(0, 1.0, 100000)
    res = uniform_to_continuous(values, 10, 15)
    assert (res >= 10).all() and (res < 15).all()
    t = stats.kstest(res, "uniform", args=(10, 5))
    assert t.pvalue > 0.4

    start, end = 1000, 500000
    res = uniform_to_continuous(values, start, end, log=True)
    assert (res >= start).all() and (res < end).all()
    t = stats.kstest(
        np.log(res), "uniform", args=(np.log(start), np.log(end) - np.log(start))
    )
    assert t.pvalue > 0.4
    t = stats.kstest(res, "uniform", args=(start, end - start))
    assert t.pvalue < 1e-3  # not uniform distribution

    res = uniform_to_continuous(values, start, end, log=True, base=10)
    assert (res >= start).all() and (res < end).all()
    b = np.log(10)
    t = stats.kstest(
        np.log(res) / b,
        "uniform",
        args=(np.log(start) / b, np.log(end) / b - np.log(start) / b),
    )
    assert t.pvalue > 0.4
    t = stats.kstest(res, "uniform", args=(start, end - start))
    assert t.pvalue < 1e-3  # not uniform distribution


def test_uniform_to_discrete():
    assert 10 == uniform_to_discrete(0.5, 10, 10, q=1.1)

    np.random.seed(0)
    values = np.random.uniform(0, 1.0, 1000)
    res = uniform_to_discrete(values, 1, 3, q=0.5)
    assert_close(res, [1.0, 1.5, 2.0, 2.5, 3.0])
    assert sum(1 if v == 3.0 else 0 for v in res) >= 180  # close to 20%
    res = uniform_to_discrete(values, 1, 3, q=0.5, include_high=False)
    assert_close(res, [1.0, 1.5, 2.0, 2.5])
    assert sum(1 if v == 2.5 else 0 for v in res) >= 230  # close to 25%
    for ih in [True, False]:
        res = uniform_to_discrete(values, 1, 3, q=0.6, include_high=ih)
        assert_close(res, [1.0, 1.6, 2.2, 2.8])
        assert sum(1 if v == 2.8 else 0 for v in res) >= 230  # close to 25%


def test_uniform_to_integers():
    assert 10 == uniform_to_integers(0.5, 10, 10)

    np.random.seed(0)
    values = np.random.uniform(0, 1.0, 1000)
    res = uniform_to_integers(values, 1, 5)
    assert all(isinstance(x, int) for x in res)
    assert set(res) == set([1, 2, 3, 4, 5])
    assert sum(1 if v == 5 else 0 for v in res) >= 180  # close to 25%
    res = uniform_to_integers(values, 1, 5, include_high=False)
    assert all(isinstance(x, int) for x in res)
    assert set(res) == set([1, 2, 3, 4])
    assert sum(1 if v == 4 else 0 for v in res) >= 230  # close to 25%

    res = uniform_to_integers(values, 1, 5, q=2)
    assert set(res) == set([1, 3, 5])
    assert sum(1 if v == 5 else 0 for v in res) >= 300  # close to 33%
    res = uniform_to_integers(values, 1, 5, q=2, include_high=False)
    assert set(res) == set([1, 3])
    assert sum(1 if v == 3 else 0 for v in res) >= 480  # close to 50%

    for ih in [True, False]:
        res = uniform_to_integers(values, 1, 5, q=3, include_high=ih)
        assert set(res) == set([1, 4])
        assert sum(1 if v == 4 else 0 for v in res) >= 480  # close to 33%


def test_uniform_to_choice():
    assert "a" == uniform_to_choice(0.5, ["a"])

    np.random.seed(0)
    values = np.random.uniform(0, 1.0, 1000)
    res = uniform_to_choice(values, ["a", "b", "c", "d"])
    for c in ["a", "b", "c", "d"]:
        assert sum(1 if v == c else 0 for v in res) >= 230  # close to 25%


def test_normal_to_continuous():
    np.random.seed(0)
    values = np.random.normal(0, 1.0, 100000)
    res = normal_to_continuous(values, 10, 15)
    t = stats.kstest(res, "norm", args=(10, 15))
    assert t.pvalue > 0.4


def test_normal_to_discrete():
    np.random.seed(0)
    values = np.random.normal(0, 1.0, 100000)
    res = normal_to_discrete(values, 0.3, 3.0, 0.0001)
    t = stats.kstest(res, "norm", args=(0.3, 3.0))
    assert t.pvalue > 0.4

    res = normal_to_discrete(values, 0.1, 0.9, q=0.3)
    res = [x for x in res if x >= -0.9 and x <= 0.9]
    assert_close([-0.8, -0.5, -0.2, 0.1, 0.4, 0.7], res)


def test_normal_to_integers():
    assert 17 == normal_to_integers(0, 17, 19)

    np.random.seed(0)
    values = np.random.normal(0, 1.0, 1000)

    res = normal_to_integers(values, 17, 19)
    assert 17 in res

    res = normal_to_integers(values, 17, 19, q=3)
    res = [x for x in res if x >= 10 and x <= 26]
    assert_close([11, 14, 17, 20, 23, 26], res)



from typing import Any, Callable, Dict, Iterable, Optional

from fugue import ArrayDataFrame, DataFrame, ExecutionEngine
from triad import assert_or_throw
from tune._utils import run_monitored_process
from tune.concepts.dataset import StudyResult, TuneDataset, get_trials_from_row
from tune.concepts.flow import RemoteTrialJudge, TrialCallback, TrialJudge, TrialReport
from tune.concepts.flow.judge import Monitor, NoOpTrailJudge
from tune.constants import TUNE_REPORT_ADD_SCHEMA, TUNE_STOPPER_DEFAULT_CHECK_INTERVAL
from tune.exceptions import TuneCompileError, TuneInterrupted
from tune.noniterative.objective import (
    NonIterativeObjectiveFunc,
    NonIterativeObjectiveLocalOptimizer,
)
from tune.noniterative.stopper import NonIterativeStopper


def _make_judge(
    monitor: Optional[Monitor] = None, stopper: Optional[NonIterativeStopper] = None
) -> Optional[TrialJudge]:
    if monitor is None and stopper is None:
        return None
    if stopper is None and monitor is not None:
        return NoOpTrailJudge(monitor)
    if stopper is not None and monitor is None:
        return stopper
    if stopper is not None and monitor is not None:
        stopper.reset_monitor(monitor)
        return stopper
    raise NotImplementedError  # pragma: no cover


class NonIterativeStudy:
    def __init__(
        self,
        objective: NonIterativeObjectiveFunc,
        optimizer: NonIterativeObjectiveLocalOptimizer,
    ):
        self._objective = objective
        self._optimizer = optimizer

    def optimize(  # noqa: C901
        self,
        dataset: TuneDataset,
        distributed: Optional[bool] = None,
        monitor: Optional[Monitor] = None,
        stopper: Optional[NonIterativeStopper] = None,
        stop_check_interval: Any = None,
    ) -> StudyResult:
        _dist = self._get_distributed(distributed)
        entrypoint: Any = None
        judge = _make_judge(monitor, stopper)
        if judge is not None:
            cb = TrialCallback(judge)
            entrypoint = cb.entrypoint
        if stopper is None:
            _interval: Any = None
        else:
            _interval = stop_check_interval or TUNE_STOPPER_DEFAULT_CHECK_INTERVAL

        def compute_processor(engine: ExecutionEngine, df: DataFrame) -> DataFrame:
            out_schema = df.schema + TUNE_REPORT_ADD_SCHEMA

            def get_rows() -> Iterable[Any]:
                for row in self._compute_transformer(
                    df.as_local().as_dict_iterable(),
                    entrypoint=entrypoint,
                    stop_check_interval=_interval,
                ):
                    yield [row[k] for k in out_schema.names]

            # TODO: need to add back execution_engine for engine aware optimizers
            # t._execution_engine = engine  # type:ignore
            return ArrayDataFrame(get_rows(), out_schema)

        def preprocess(df: DataFrame) -> DataFrame:
            if judge is not None:
                judge.monitor.initialize()
            return df

        def postprocess(df: DataFrame) -> None:
            if judge is not None:
                judge.monitor.finalize()

        if not _dist:
            res = dataset.data.process(preprocess).process(compute_processor)
        else:
            res = (
                dataset.data.process(preprocess)
                .per_row()
                .transform(
                    self._compute_transformer,
                    schema=f"*,{TUNE_REPORT_ADD_SCHEMA}",
                    callback=entrypoint,
                    params=dict(stop_check_interval=_interval),
                )
            )

        res.persist().output(postprocess)

        return StudyResult(dataset=dataset, result=res)

    def _get_distributed(self, distributed: Optional[bool]) -> bool:
        if distributed is None:
            return self._optimizer.distributable
        if distributed:
            assert_or_throw(
                self._optimizer.distributable,
                TuneCompileError(
                    f"can't distribute non-distributable optimizer {self._optimizer}"
                ),
            )
            return True
        return False

    def _compute_transformer(
        self,
        df: Iterable[Dict[str, Any]],
        entrypoint: Optional[Callable[[str, Dict[str, Any]], Any]] = None,
        stop_check_interval: Any = None,
    ) -> Iterable[Dict[str, Any]]:
        j: Optional[RemoteTrialJudge] = (
            None if entrypoint is None else RemoteTrialJudge(entrypoint)
        )
        for row in df:
            for n, trial in enumerate(get_trials_from_row(row, with_dfs=False)):
                if j is not None:
                    if stop_check_interval is None:
                        # monitor only
                        report = self._local_process_trial(row, n)
                        j.judge(report)
                        yield report.fill_dict(dict(row))
                    elif j.can_accept(trial):
                        try:
                            report = run_monitored_process(
                                self._local_process_trial,
                                [row, n],
                                {},
                                lambda: not j.can_accept(trial),  # type: ignore
                                stop_check_interval,
                            )
                        except TuneInterrupted:
                            continue
                        except Exception:
                            raise
                        j.judge(report)
                        yield report.fill_dict(dict(row))
                else:
                    report = self._local_process_trial(row, n)
                    yield report.fill_dict(dict(row))

    def _local_process_trial(self, row: Dict[str, Any], idx: int) -> TrialReport:
        trial = list(get_trials_from_row(row))[idx]
        report = self._optimizer.run(self._objective, trial)
        return report.with_sort_metric(
            self._objective.generate_sort_metric(report.metric)
        )



from typing import Any, Dict, Iterable, List, Optional

import numpy as np
from triad import assert_or_throw, to_uuid
from tune._utils.math import (
    normal_to_continuous,
    normal_to_discrete,
    normal_to_integers,
    uniform_to_continuous,
    uniform_to_discrete,
    uniform_to_integers,
)


class Grid(object):
    """Grid search, every value will be used

    :param args: values for the grid search
    """

    def __init__(self, *args: Any):
        assert_or_throw(
            len(args) > 0, ValueError("Grid must take at least one element")
        )
        self._values = list(args)

    def __iter__(self) -> Iterable[Any]:
        yield from self._values

    def __eq__(self, other: Any):
        """Compare two ``Grid``"""
        return isinstance(other, type(self)) and self._values == other._values

    def __uuid__(self) -> str:
        return to_uuid("grid", self._values)


class StochasticExpression(object):
    """Stochastic search base class"""

    @property
    def jsondict(self) -> Dict[str, Any]:  # pragma: no cover
        """Dict representation of the expression that is json serializable"""
        raise NotImplementedError

    def __eq__(self, other: Any):
        """Compare two ``StochasticExpression``"""
        return isinstance(other, type(self)) and self.jsondict == other.jsondict

    def generate(self, seed: Any = None) -> Any:  # pragma: no cover
        """Return a randomly chosen value.

        :param seed: if set, it will be used to call :func:`~np:numpy.random.seed`
          , defaults to None
        """
        raise NotImplementedError

    def generate_many(self, n: int, seed: Any = None) -> List[Any]:
        if seed is not None:
            np.random.seed(seed)
        return [self.generate() for _ in range(n)]

    def __uuid__(self) -> str:
        return to_uuid(self.jsondict)


class Choice(StochasticExpression):
    """A random choice of values

    :param args: values to choose from
    """

    def __init__(self, *args: Any):
        assert_or_throw(
            len(args) > 0, ValueError("Choice must take at least one element")
        )
        self._values = list(args)

    @property
    def values(self) -> List[Any]:
        """values to choose from"""
        return self._values

    @property
    def jsondict(self) -> Dict[str, Any]:
        return dict(_expr_="choice", values=self.values)

    def generate(self, seed: Any = None) -> Any:
        if seed is not None:
            np.random.seed(seed)
        value = np.random.choice(self._values)
        if isinstance(value, np.generic):
            return value.item()
        return value  # pragma: no cover


class RandBase(StochasticExpression):
    """Base class for continuous random variables

    :param q: step between adjacent values, if set, the value will be rounded
      using ``q``, defaults to None
    :param log: whether to apply ``exp`` to the random variable so that the logarithm
      of the return value is uniformly distributed, defaults to False
    """

    def __init__(
        self,
        q: Optional[float] = None,
        log: bool = False,
    ):
        if q is not None:
            assert_or_throw(q > 0, f"{q} <= 0")
        self.q = q
        self.log = log


class Rand(RandBase):
    """Continuous uniform random variables

    :param low: range low bound (inclusive)
    :param high: range high bound (exclusive)
    :param q: step between adjacent values, if set, the value will be rounded
      using ``q``, defaults to None
    :param log: whether to apply ``exp`` to the random variable so that the logarithm
      of the return value is uniformly distributed, defaults to False
    """

    def __init__(
        self,
        low: float,
        high: float,
        q: Optional[float] = None,
        log: bool = False,
        include_high: bool = True,
    ):
        if include_high:
            assert_or_throw(high >= low, ValueError(f"{high} < {low}"))
        else:
            assert_or_throw(high > low, ValueError(f"{high} <= {low}"))
        assert_or_throw(q is None or q > 0, ValueError(q))
        if log:
            assert_or_throw(
                low >= 1.0,
                ValueError(
                    f"for log sampling, low ({low}) must be greater or equal to 1.0"
                ),
            )
        self.low = low
        self.high = high
        self.include_high = include_high
        super().__init__(q, log)

    @property
    def jsondict(self) -> Dict[str, Any]:
        res = dict(
            _expr_="rand",
            low=self.low,
            high=self.high,
            log=self.log,
            include_high=self.include_high,
        )
        if self.q is not None:
            res["q"] = self.q
        return res

    def generate(self, seed: Any = None) -> float:
        if seed is not None:
            np.random.seed(seed)
        value = np.random.uniform()
        if self.q is None:
            return float(
                uniform_to_continuous(value, self.low, self.high, log=self.log)
            )
        else:
            return float(
                uniform_to_discrete(
                    value,
                    self.low,
                    self.high,
                    q=self.q,
                    log=self.log,
                    include_high=self.include_high,
                )
            )


class RandInt(RandBase):
    """Uniform distributed random integer values

    :param low: range low bound (inclusive)
    :param high: range high bound (exclusive)
    :param log: whether to apply ``exp`` to the random variable so that the logarithm
      of the return value is uniformly distributed, defaults to False
    """

    def __init__(
        self,
        low: int,
        high: int,
        q: int = 1,
        log: bool = False,
        include_high: bool = True,
    ):
        if include_high:
            assert_or_throw(high >= low, ValueError(f"{high} < {low}"))
        else:
            assert_or_throw(high > low, ValueError(f"{high} <= {low}"))
        assert_or_throw(q > 0, ValueError(q))
        if log:
            assert_or_throw(
                low >= 1.0,
                ValueError(
                    f"for log sampling, low ({low}) must be greater or equal to 1.0"
                ),
            )
        self.low = low
        self.high = high
        self.include_high = include_high
        super().__init__(q, log)

    @property
    def jsondict(self) -> Dict[str, Any]:
        return dict(
            _expr_="randint",
            low=self.low,
            high=self.high,
            q=self.q,
            log=self.log,
            include_high=self.include_high,
        )

    def generate(self, seed: Any = None) -> float:
        if seed is not None:
            np.random.seed(seed)
        value = np.random.uniform()
        return int(
            uniform_to_integers(  # type: ignore
                value,
                self.low,
                self.high,
                q=int(self.q),  # type: ignore
                log=self.log,
                include_high=self.include_high,
            )
        )


class NormalRand(RandBase):
    """Continuous normally distributed random variables

    :param mu: mean of the normal distribution
    :param sigma: standard deviation of the normal distribution
    :param q: step between adjacent values, if set, the value will be rounded
      using ``q``, defaults to None
    :param log: whether to apply ``exp`` to the random variable so that the logarithm
      of the return value is uniformly distributed, defaults to False
    """

    def __init__(
        self,
        mu: float,
        sigma: float,
        q: Optional[float] = None,
    ):
        assert_or_throw(sigma > 0, ValueError(sigma))
        assert_or_throw(q is None or q > 0, ValueError(q))
        self.mu = mu
        self.sigma = sigma
        super().__init__(q)

    @property
    def jsondict(self) -> Dict[str, Any]:
        res = dict(
            _expr_="randn",
            mu=self.mu,
            sigma=self.sigma,
        )
        if self.q is not None:
            res["q"] = self.q
        return res

    def generate(self, seed: Any = None) -> float:
        if seed is not None:
            np.random.seed(seed)
        value = np.random.normal()
        if self.q is not None:
            return normal_to_discrete(value, mean=self.mu, sigma=self.sigma, q=self.q)
        else:
            return normal_to_continuous(value, mean=self.mu, sigma=self.sigma)


class NormalRandInt(RandBase):
    """Normally distributed random integer values

    :param mu: mean of the normal distribution
    :param sigma: standard deviation of the normal distribution
    :param log: whether to apply ``exp`` to the random variable so that the logarithm
      of the return value is uniformly distributed, defaults to False
    """

    def __init__(
        self,
        mu: int,
        sigma: float,
        q: int = 1,
    ):
        assert_or_throw(sigma > 0, ValueError(sigma))
        assert_or_throw(q > 0, ValueError(q))
        self.mu = mu
        self.sigma = sigma
        self.q = q
        super().__init__(q)

    @property
    def jsondict(self) -> Dict[str, Any]:
        return dict(
            _expr_="randnint",
            mu=self.mu,
            sigma=self.sigma,
            q=self.q,
        )

    def generate(self, seed: Any = None) -> int:
        if seed is not None:
            np.random.seed(seed)
        value = np.random.normal()
        return int(
            normal_to_integers(  # type: ignore
                value,
                mean=self.mu,
                sigma=self.sigma,
                q=self.q,  # type: ignore
            )
        )


def encode_params(value: Any):
    if isinstance(value, StochasticExpression):
        return value.jsondict
    elif isinstance(value, str):
        return value
    elif isinstance(value, list):
        return [encode_params(v) for v in value]
    elif isinstance(value, dict):
        return {k: encode_params(v) for k, v in value.items()}
    return value


def decode_params(value: Any) -> Any:
    if isinstance(value, str):
        return value
    elif isinstance(value, list):
        return [decode_params(v) for v in value]
    elif isinstance(value, dict):
        if "_expr_" in value:
            e = value.pop("_expr_")
            if e == "choice":  # TODO: embeded rand is not supported
                return Choice(*value["values"])
            if e == "rand":
                return Rand(**value)
            if e == "randint":
                return RandInt(**value)
            if e == "randn":
                return NormalRand(**value)
            if e == "randnint":
                return NormalRandInt(**value)
            raise ValueError(e)  # pragma: no cover
        else:
            return {k: decode_params(v) for k, v in value.items()}
    else:
        return value



from threading import RLock
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple

from triad import to_uuid
from tune.concepts.flow import (
    Monitor,
    Trial,
    TrialDecision,
    TrialJudge,
    TrialReport,
    TrialReportHeap,
)


class RungHeap:
    def __init__(self, n: int):
        self._lock = RLock()
        self._n = n
        self._heap = TrialReportHeap(min_heap=False)
        self._bests: List[float] = []

    def __len__(self) -> int:
        with self._lock:
            return len(self._heap)

    @property
    def capacity(self) -> int:
        return self._n

    @property
    def best(self) -> float:
        with self._lock:
            return self._bests[-1] if len(self._bests) > 0 else float("nan")

    @property
    def bests(self) -> List[float]:
        with self._lock:
            return self._bests

    @property
    def full(self) -> bool:
        with self._lock:
            return self.capacity <= len(self)

    def __contains__(self, tid: str) -> bool:
        with self._lock:
            return tid in self._heap

    def values(self) -> Iterable[TrialReport]:
        return self._heap.values()

    def push(self, report: TrialReport) -> bool:
        with self._lock:
            if len(self) == 0:
                best = report.sort_metric
            else:
                best = min(self.best, report.sort_metric)
            self._heap.push(report)
            self._bests.append(best)
            return (
                len(self._heap) <= self._n
                or self._heap.pop().trial_id != report.trial_id
            )


class _PerTrial:
    def __init__(self, parent: "_PerPartition") -> None:
        self._history: List[TrialReport] = []
        self._parent = parent
        self._active = True

    def can_promote(self, report: TrialReport) -> Tuple[bool, str]:
        reasons: List[str] = []
        if self._active:
            can_accept = self._parent.can_accept(report.trial)
            early_stop = self._parent._parent._trial_early_stop(
                report, self._history, self._parent._rungs
            )
            self._active = can_accept and not early_stop
            if not can_accept:
                reasons.append("can't accept new")
            if early_stop:
                reasons.append("trial early stop")
        if self._active:
            self._history.append(report)
            can_push = self._parent._rungs[report.rung].push(report)
            if not can_push:
                # data = sorted(
                #     (x.jsondict for x in self._parent._rungs[report.rung].values()),
                #     key=lambda x: x["sort_metric"],
                # )
                # reasons.append("not best: " + json.dumps(data))
                reasons.append("not best")
            return can_push, ", ".join(reasons)
        return False, ", ".join(reasons)

    def judge(self, report: TrialReport) -> TrialDecision:
        if report.rung >= len(self._parent._parent.schedule) - 1:
            self._history.append(report)
            self._parent._rungs[report.rung].push(report)
            return TrialDecision(
                report, budget=0, should_checkpoint=True, reason="last"
            )
        promote, reason = self.can_promote(report)
        if not promote:
            return TrialDecision(
                report, budget=0, should_checkpoint=True, reason=reason
            )
        next_budget = self._parent.get_budget(report.trial, report.rung + 1)
        return TrialDecision(
            report,
            budget=next_budget,
            should_checkpoint=next_budget <= 0
            or self._parent._parent.always_checkpoint,
            reason="" if next_budget > 0 else "budget==0",
        )


class _PerPartition:
    def __init__(self, parent: "ASHAJudge", keys: List[Any]):
        self._keys = keys
        self._data: Dict[str, _PerTrial] = {}
        self._lock = RLock()
        self._parent = parent
        self._rungs: List[RungHeap] = [RungHeap(x[1]) for x in self._parent.schedule]
        self._active = True
        self._accepted_ids: Set[str] = set()

    def can_accept(self, trial: Trial) -> bool:
        with self._lock:
            if self._active:
                self._active = not self._parent._study_early_stop(
                    self._keys, self._rungs
                )
                if self._active:
                    self._accepted_ids.add(trial.trial_id)
                    return True
            # if not active, can only accept existing trials
            return trial.trial_id in self._accepted_ids

    def get_budget(self, trial: Trial, rung: int) -> float:
        if rung >= len(self._parent.schedule) or not self.can_accept(trial):
            return 0.0  # pragma: no cover
        return self._parent.schedule[rung][0]

    def judge(self, report: TrialReport) -> TrialDecision:
        return self._get_judge(report.trial).judge(report)

    def _get_judge(self, trial: Trial) -> _PerTrial:
        key = trial.trial_id
        with self._lock:
            if key not in self._data:
                self._data[key] = _PerTrial(self)
            return self._data[key]


class ASHAJudge(TrialJudge):
    def __init__(
        self,
        schedule: List[Tuple[float, int]],
        always_checkpoint: bool = False,
        study_early_stop: Optional[Callable[[List[Any], List[RungHeap]], bool]] = None,
        trial_early_stop: Optional[
            Callable[[TrialReport, List[TrialReport], List[RungHeap]], bool]
        ] = None,
        monitor: Optional[Monitor] = None,
    ):
        super().__init__(monitor=monitor)
        self._lock = RLock()
        self._data: Dict[str, _PerPartition] = {}
        self._schedule = schedule
        self._always_checkpoint = always_checkpoint
        self._study_early_stop = study_early_stop or _default_study_early_stop
        self._trial_early_stop = trial_early_stop or _default_trial_early_stop

    @property
    def schedule(self) -> List[Tuple[float, int]]:
        return self._schedule

    @property
    def always_checkpoint(self) -> bool:
        return self._always_checkpoint

    def can_accept(self, trial: Trial) -> bool:
        return self._get_judge(trial).can_accept(trial)

    def get_budget(self, trial: Trial, rung: int) -> float:
        budget = self._get_judge(trial).get_budget(trial, rung)
        self.monitor.on_get_budget(trial, rung, budget)
        return budget

    def judge(self, report: TrialReport) -> TrialDecision:
        self.monitor.on_report(report)
        decision = self._get_judge(report.trial).judge(report)
        self.monitor.on_judge(decision)
        return decision

    def _get_judge(self, trial: Trial) -> _PerPartition:
        key = to_uuid(trial.keys)
        with self._lock:
            if key not in self._data:
                self._data[key] = _PerPartition(self, trial.keys)
            return self._data[key]


def _default_study_early_stop(keys: List[Any], rungs: List["RungHeap"]) -> bool:
    return all(r.full for r in rungs)


def _default_trial_early_stop(
    report: TrialReport, reports: List[TrialReport], rungs: List["RungHeap"]
) -> bool:
    return False



from typing import Any, Dict, Type

from sklearn.base import is_classifier, is_regressor
from triad import assert_or_throw
from triad.utils.convert import get_full_type_path, to_type
from tune.constants import SPACE_MODEL_NAME
from tune.concepts.space.spaces import Space


def to_sk_model(obj: Any) -> Type:
    if isinstance(obj, str):
        obj = to_type(obj)
    assert_or_throw(
        is_classifier(obj) or is_regressor(obj),
        TypeError(f"{obj} is neither a sklearn classifier or regressor"),
    )
    return obj


def to_sk_model_expr(model: Any) -> Any:
    if isinstance(model, str):
        model = to_sk_model(model)
    return get_full_type_path(model)


def sk_space(model: str, **params: Dict[str, Any]) -> Space:
    data = {SPACE_MODEL_NAME: to_sk_model_expr(model)}
    data.update(params)
    return Space(**data)



from tune._utils import _EMPTY_ITER, dict_product, product, safe_iter


def test_safe_iter():
    assert [1] == list(safe_iter([1]))
    assert [1] == list(safe_iter(safe_iter([1])))
    assert [None] == list(safe_iter(safe_iter([None])))
    assert [1] == list(safe_iter([1], safe=False))

    assert [_EMPTY_ITER] == list(safe_iter([]))
    assert [_EMPTY_ITER] == list(safe_iter(safe_iter([])))
    assert [] == list(safe_iter([], safe=False))


def test_product():
    res = list(product([], safe=False))
    assert 0 == len(res)
    res = list(product([], safe=True))
    assert 1 == len(res)

    res = list(product([[]], safe=False))
    assert 0 == len(res)
    res = list(product([[]], safe=True))
    assert [[]] == res

    res = list(product([[], []]))
    assert 0 == len(res)
    res = list(product([[], []], safe=True))
    assert [[]] == res
    res = list(product([[], []], safe=True, remove_empty=False))
    assert [[None, None]] == res

    for safe in [True, False]:
        for remove_empty in [True, False]:
            res = list(product([[1], [2]], safe=safe, remove_empty=remove_empty))
            assert 1 == len(res)
            assert [1, 2] == res[0]

            res = list(product([[1], [2, 3]], safe=safe, remove_empty=remove_empty))
            assert [[1, 2], [1, 3]] == res

    res = list(product([[1], []]))
    assert [] == res
    res = list(product([[None], [], [2, 3]], safe=True, remove_empty=False))
    assert [[None, None, 2], [None, None, 3]] == res
    res = list(product([[None], [], [2, 3]], safe=True, remove_empty=True))
    assert [[None, 2], [None, 3]] == res

    res = list(product([[1], safe_iter([])], safe=False, remove_empty=True))
    assert [[1]] == res
    res = list(product([[1], safe_iter([])], safe=False, remove_empty=False))
    assert [[1, None]] == res


def test_dict_product():
    res = list(dict_product({}, safe=True))
    assert [{}] == res
    res = list(dict_product({"a": []}, safe=True))
    assert [{}] == res

    res = list(dict_product({}, safe=False))
    assert [] == res
    res = list(dict_product({"a": []}, safe=False))
    assert [] == res



import copy
import inspect
from typing import Any, Callable, Dict, Optional, Tuple, no_type_check

from fugue._utils.interfaceless import (
    FunctionWrapper,
    _FuncParam,
    _OtherParam,
    is_class_method,
)
from triad import assert_or_throw
from triad.utils.convert import get_caller_global_local_vars, to_function
from tune.exceptions import TuneCompileError
from tune.noniterative.objective import NonIterativeObjectiveFunc
from tune.concepts.flow import Trial, TrialReport


def noniterative_objective(
    func: Optional[Callable] = None, min_better: bool = True
) -> Callable[[Any], NonIterativeObjectiveFunc]:
    def deco(func: Callable) -> NonIterativeObjectiveFunc:
        assert_or_throw(
            not is_class_method(func),
            NotImplementedError(
                "non_iterative_objective decorator can't be used on class methods"
            ),
        )
        return _NonIterativeObjectiveFuncWrapper.from_func(func, min_better)

    if func is None:
        return deco
    else:
        return deco(func)  # type: ignore


def to_noniterative_objective(
    obj: Any,
    min_better: bool = True,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> NonIterativeObjectiveFunc:
    if isinstance(obj, NonIterativeObjectiveFunc):
        return copy.copy(obj)
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, NonIterativeObjectiveFunc):
            return copy.copy(f)
        # this is for functions without decorator
        return _NonIterativeObjectiveFuncWrapper.from_func(f, min_better)
    except Exception as e:
        exp = e
    raise TuneCompileError(f"{obj} is not a valid tunable function", exp)


class _NonIterativeObjectiveFuncWrapper(NonIterativeObjectiveFunc):
    def __init__(self, min_better: bool):
        self._min_better = min_better

    @property
    def min_better(self) -> bool:
        return self._min_better

    def generate_sort_metric(self, value: float) -> float:
        return float(value) if self._min_better else -float(value)

    @no_type_check
    def run(self, trial: Trial) -> TrialReport:
        if self._orig_input:
            result = self._func(trial)
        else:
            result = self._func(**trial.params, **trial.dfs)
        return self._output_f(result, trial)

    @no_type_check
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._func(*args, **kwargs)

    @no_type_check
    @staticmethod
    def from_func(
        func: Callable, min_better: bool
    ) -> "_NonIterativeObjectiveFuncWrapper":
        f = _NonIterativeObjectiveFuncWrapper(min_better=min_better)
        w = _NonIterativeObjectiveWrapper(func)
        f._func = w._func
        f._orig_input = w._orig_input
        f._output_f = w._rt.to_report
        return f


class _ReportParam(_FuncParam):
    def __init__(self, param: Optional[inspect.Parameter], annotation: Any, code: str):
        super().__init__(param, annotation, code)

    def to_report(self, v: Any, trial: Trial) -> TrialReport:
        raise NotImplementedError  # pragma: no cover


class _RawReportParam(_ReportParam):
    def __init__(self, param: Optional[inspect.Parameter]):
        super().__init__(param, "TrialReport", "r")

    def to_report(self, v: Any, trial: Trial) -> TrialReport:
        return v


class _MetricParam(_ReportParam):
    def __init__(self, param: Optional[inspect.Parameter]):
        super().__init__(param, "float", "1")

    def to_report(self, v: Any, trial: Trial) -> TrialReport:
        return TrialReport(trial, metric=float(v), params=trial.params, metadata={})


class _MetricMetadataParam(_ReportParam):
    def __init__(self, param: Optional[inspect.Parameter]):
        super().__init__(param, "Tuple[float,Dict[str,Any]]", "2")

    def to_report(self, v: Any, trial: Trial) -> TrialReport:
        return TrialReport(
            trial, metric=float(v[0]), params=trial.params, metadata=v[1]
        )


class _TrialParam(_FuncParam):
    def __init__(self, param: Optional[inspect.Parameter]):
        super().__init__(param, "Trial", "t")


class _NonIterativeObjectiveWrapper(FunctionWrapper):
    def __init__(self, func: Callable):
        self._orig_input = False
        self._orig_output = False
        super().__init__(func, "^(t|([x1]+))$", "^[r12]$")

    def _parse_param(
        self,
        annotation: Any,
        param: Optional[inspect.Parameter],
        none_as_other: bool = True,
    ) -> _FuncParam:
        if annotation is float:
            return _MetricParam(param)
        elif annotation is Tuple[float, Dict[str, Any]]:
            return _MetricMetadataParam(param)
        elif annotation is TrialReport:
            self._orig_output = True
            return _RawReportParam(param)
        elif annotation is Trial:
            self._orig_input = True
            return _TrialParam(param)
        else:
            return _OtherParam(param)



# flake8: noqa

from tune_tensorflow.objective import KerasObjective
from tune_tensorflow.spec import KerasTrainingSpec
from tune_tensorflow.suggest import (
    suggest_keras_models_by_continuous_asha,
    suggest_keras_models_by_hyperband,
    suggest_keras_models_by_sha,
)
from tune_tensorflow.utils import keras_space



import heapq
import json
from datetime import datetime
from threading import RLock
from typing import Any, Dict, Iterable, List, Optional, Set

from triad.utils.convert import to_datetime
from tune.concepts.space.parameters import decode_params, encode_params
from tune.concepts.flow.trial import Trial
from tune.constants import TUNE_REPORT, TUNE_REPORT_ID, TUNE_REPORT_METRIC


class TrialReport:
    def __init__(
        self,
        trial: Trial,
        metric: Any,
        params: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        cost: float = 1.0,
        rung: int = 0,
        sort_metric: Any = None,
        raw: bool = False,
        log_time: Any = None,
    ):
        self._trial = trial.with_dfs({})
        self._metric = float(metric)
        self._cost = float(cost)
        self._rung = rung
        self._sort_metric = self._metric if sort_metric is None else float(sort_metric)
        if params is None:
            self._params = trial.params
        else:
            self._params = params if raw else decode_params(params)
        self._metadata = metadata or {}
        self._log_time = datetime.now() if log_time is None else to_datetime(log_time)

    def copy(self) -> "TrialReport":
        return TrialReport(
            trial=self.trial,
            metric=self.metric,
            params=self._params,
            metadata=self._metadata,
            cost=self._cost,
            rung=self._rung,
            sort_metric=self._sort_metric,
            raw=True,
            log_time=self.log_time,
        )

    def __copy__(self) -> "TrialReport":
        return self.copy()

    def __deepcopy__(self, memo: Any) -> "TrialReport":
        return self.copy()

    @property
    def log_time(self) -> datetime:
        return self._log_time

    def reset_log_time(self) -> "TrialReport":
        res = self.copy()
        res._log_time = datetime.now()
        return res

    @property
    def trial(self) -> Trial:
        return self._trial

    @property
    def trial_id(self) -> str:
        return self.trial.trial_id

    @property
    def metric(self) -> float:
        return self._metric

    @property
    def sort_metric(self) -> float:
        return self._sort_metric

    @property
    def cost(self) -> float:
        return self._cost

    @property
    def rung(self) -> int:
        return self._rung

    @property
    def params(self) -> Dict[str, Any]:
        return self._params

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    def with_cost(self, cost: float) -> "TrialReport":
        t = self.copy()
        t._cost = cost
        return t

    def with_rung(self, rung: int) -> "TrialReport":
        t = self.copy()
        t._rung = rung
        return t

    def with_sort_metric(self, sort_metric: Any) -> "TrialReport":
        t = self.copy()
        t._sort_metric = float(sort_metric)
        return t

    def generate_sort_metric(self, min_better: bool, digits: int) -> "TrialReport":
        t = self.copy()
        t._sort_metric = (
            round(self.metric, digits) if min_better else -round(self.metric, digits)
        )
        return t

    @property
    def jsondict(self) -> Dict[str, Any]:
        return {
            "trial": self.trial.jsondict,
            "metric": self.metric,
            "params": encode_params(self.params),
            "metadata": self.metadata,
            "cost": self.cost,
            "rung": self.rung,
            "sort_metric": self.sort_metric,
            "log_time": str(self.log_time),
        }

    @staticmethod
    def from_jsondict(data: Dict[str, Any]) -> "TrialReport":
        trial = Trial.from_jsondict(data.pop("trial"))
        return TrialReport(trial=trial, **data)

    def fill_dict(self, data: Dict[str, Any]) -> Dict[str, Any]:
        data[TUNE_REPORT_ID] = self.trial_id
        data[TUNE_REPORT_METRIC] = self.sort_metric
        data[TUNE_REPORT] = json.dumps(self.jsondict)
        return data


class TrialReportHeap:
    class _Wrapper:
        def __init__(self, report: TrialReport, min_heap: bool):
            self.report = report
            self.min_heap = min_heap

        def __lt__(self, other: "TrialReportHeap._Wrapper") -> bool:
            k1 = (self.report.sort_metric, self.report.cost, self.report.rung)
            k2 = (other.report.sort_metric, other.report.cost, other.report.rung)
            return k1 < k2 if self.min_heap else k2 < k1

        @property
        def trial_id(self) -> str:
            return self.report.trial_id

    def __init__(self, min_heap: bool):
        self._data: List[TrialReportHeap._Wrapper] = []
        self._ids: Set[str] = set()
        self._min_heap = min_heap

    def __len__(self) -> int:
        return len(self._data)

    def __contains__(self, tid: str) -> bool:
        return tid in self._ids

    def values(self) -> Iterable[TrialReport]:
        for x in self._data:
            yield x.report

    def push(self, report: TrialReport) -> None:
        w = TrialReportHeap._Wrapper(report, self._min_heap)
        if w.trial_id in self._ids:
            self._data = [x if x.trial_id != w.trial_id else w for x in self._data]
            heapq.heapify(self._data)
        else:
            self._ids.add(w.trial_id)
            heapq.heappush(self._data, w)

    def pop(self) -> TrialReport:
        w = heapq.heappop(self._data)
        self._ids.remove(w.trial_id)
        return w.report


class TrialReportLogger:
    def __init__(self, new_best_only: bool = False):
        self._lock = RLock()
        self._best_report: Optional[TrialReport] = None
        self._new_best_only = new_best_only

    def on_report(self, report: TrialReport) -> bool:
        with self._lock:
            updated = False
            if (
                self._best_report is None
                or report.sort_metric < self._best_report.sort_metric
            ):
                self._best_report = report
                updated = True
            if updated or not self._new_best_only:
                self.log(report)
            return updated

    def log(self, report: TrialReport) -> None:  # pragma: no cover
        pass

    @property
    def best(self) -> Optional[TrialReport]:
        with self._lock:
            return self._best_report



import math
from typing import Any, Dict, Iterable

from fugue import FugueWorkflow
from tune import optimize_by_continuous_asha
from tune.constants import TUNE_REPORT_METRIC
from tune.concepts.dataset import TuneDatasetBuilder
from tune.iterative.asha import ASHAJudge, RungHeap
from tune.iterative.objective import IterativeObjectiveFunc
from tune.concepts.space import Grid, Space
from tune.concepts.flow import Monitor, Trial, TrialReport


def test_rung_heap():
    h = RungHeap(2)
    assert 2 == h.capacity
    assert not h.full
    assert math.isnan(h.best)
    assert 0 == len(h)
    assert [] == h.bests
    assert h.push(rp("a", 1))
    assert h.push(rp("b", 0.1))
    assert h.push(rp("a", 2))
    assert h.push(rp("c", 0.5))
    assert not h.push(rp("d", 0.6))
    assert 0.1 == h.best
    assert 2 == len(h)
    assert h.full
    assert "a" not in h
    assert "d" not in h
    assert "b" in h
    assert "c" in h
    assert [1.0, 0.1, 0.1, 0.1, 0.1] == h.bests
    assert h.push(rp("e", 0.01))
    assert [1.0, 0.1, 0.1, 0.1, 0.1, 0.01] == h.bests
    assert "b" in h and "e" in h and 2 == len(h)
    assert h.push(rp("e", 5))
    assert [1.0, 0.1, 0.1, 0.1, 0.1, 0.01, 0.01] == h.bests
    values = {x.trial_id: x for x in h.values()}
    assert "b" in values and "e" in values and 2 == len(values)
    assert 5.0 == values["e"].sort_metric


# TODO: test multiple partitions


def test_asha_judge_simple_happy_path():
    j = ASHAJudge(schedule=[(1.0, 2), (2.0, 1)])
    d = j.judge(rp("a", 0.5, 0))
    assert 2.0 == d.budget
    assert not d.should_checkpoint
    d = j.judge(rp("b", 0.6, 0))
    assert 2.0 == d.budget
    assert not d.should_checkpoint
    d = j.judge(rp("a", 0.4, 1))
    assert d.should_stop
    assert d.should_checkpoint
    # stop criteria met, so other jobs won't get more budget
    d = j.judge(rp("c", 0.2, 0))
    assert d.should_stop
    assert d.should_checkpoint


def test_asha_stop():
    def should_stop(keys, rungs):
        metrics = []
        for r in rungs:
            if len(r) == 0:
                break
            if not r.full:
                return False
            metrics.append(r.best)
        if len(metrics) < 2:
            return False
        return metrics[-2] - metrics[-1] < 0.2

    j = ASHAJudge(
        schedule=[(1.0, 2), (2.0, 2), (3.0, 1)],
        always_checkpoint=True,
        study_early_stop=should_stop,
    )
    d = j.judge(rp("a", 0.6, 0))
    assert 2.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("b", 0.5, 0))
    assert 2.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("c", 0.4, 0))
    assert 2.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("b", 0.45, 1))
    assert 3.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("c", 0.39, 1))
    assert 3.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("x", 0.45, 0))
    # rungs[1] and rungs[0] diff so somall
    # no longer accept new trials
    assert 0.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("b", 0.45, 1))
    assert 3.0 == d.budget  # existed ids can still be accepted
    assert d.should_checkpoint
    d = j.judge(rp("a", 0.44, 2))
    assert 0.0 == d.budget  # already stopped
    assert d.should_checkpoint


def test_trial_stop():
    def should_stop(report, history, rungs):
        return not all(report.trial_id in x for x in rungs[: report.rung])

    j = ASHAJudge(
        schedule=[(1.0, 2), (2.0, 2), (3.0, 1)],
        always_checkpoint=True,
        trial_early_stop=should_stop,
    )
    d = j.judge(rp("a", 0.6, 0))
    assert 2.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("b", 0.5, 0))
    assert 2.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("c", 0.4, 0))
    assert 2.0 == d.budget
    assert d.should_checkpoint
    d = j.judge(rp("a", 0.1, 1))
    assert d.should_stop  # kicked out by c
    assert d.should_checkpoint


def test_run_asha(tmpdir):
    class M(Monitor):
        def on_report(self, report: TrialReport) -> None:
            print(report.jsondict)

    def assert_metric(df: Iterable[Dict[str, Any]], metric: float, ct: int) -> None:
        n = 0
        for row in df:
            assert row[TUNE_REPORT_METRIC] == metric
            n += 1
        assert n == ct

    space = Space(a=Grid(0, 1, 2, 3))
    dag = FugueWorkflow()
    dataset = TuneDatasetBuilder(space, str(tmpdir)).build(dag, shuffle=False)
    obj = F()
    res = optimize_by_continuous_asha(
        obj,
        dataset,
        plan=[[1.0, 3], [1.0, 2], [1.0, 1], [1.0, 1]],
        checkpoint_path=str(tmpdir),
    )
    res.result(1).output(assert_metric, dict(metric=1.0, ct=1))

    res = optimize_by_continuous_asha(
        obj,
        dataset,
        plan=[[2.0, 2], [1.0, 1], [1.0, 1]],
        checkpoint_path=str(tmpdir),
        monitor=M(),
    )
    res.result(1).output(assert_metric, dict(metric=1.0, ct=1))
    dag.run()


def rp(tid, metric, rung=0, keys=[]):
    t = Trial(tid, {}, keys=keys)
    return TrialReport(t, metric=metric, rung=rung)


class F(IterativeObjectiveFunc):
    def __init__(self):
        super().__init__()
        self._it = 0
        self._all = [
            [9, 3, 1, 1],
            [8, 6, 5, 5],
            [8, 5, 4, 3],
            [7, 4, 3, 4],
        ]

    def save_checkpoint(self, fs):
        fs.writetext("x", str(self._it))

    def load_checkpoint(self, fs):
        self._it = int(fs.readtext("x"))

    def run_single_iteration(self):
        trial = self.current_trial
        metric = self._all[trial.params["a"]][self._it]
        self._it += 1
        return TrialReport(trial, metric=metric)

    def copy(self):
        return F()



from pytest import raises
from tune import TUNE_OBJECT_FACTORY, RandInt
from tune.exceptions import TuneCompileError

from tune_tensorflow import (
    keras_space,
    suggest_keras_models_by_continuous_asha,
    suggest_keras_models_by_hyperband,
    suggest_keras_models_by_sha,
)

from tests.tune_tensorflow.mock import MockSpec


def test_sha(tmpdir):
    TUNE_OBJECT_FACTORY.set_temp_path(str(tmpdir))

    space = keras_space(MockSpec, l1=RandInt(8, 16), l2=RandInt(8, 24))
    with raises(TuneCompileError):
        suggest_keras_models_by_sha(space, plan=[(2.0, 4), (4.0, 2)])

    space = space.sample(6, 0)
    reports = suggest_keras_models_by_sha(space, plan=[(2.0, 4), (4.0, 2)], top_n=2)
    for r in reports:
        print(r.jsondict)
    assert 2 == len(reports)


def test_hyperband(tmpdir):
    TUNE_OBJECT_FACTORY.set_temp_path(str(tmpdir))

    space = keras_space(MockSpec, l1=RandInt(8, 16), l2=RandInt(8, 24))
    with raises(TuneCompileError):
        suggest_keras_models_by_hyperband(
            space,
            plans=[
                [(2.0, 4), (4.0, 2)],
                [(4.0, 2), (2.0, 4)],
            ],
        )

    space = space.sample(10, 0)
    reports = suggest_keras_models_by_hyperband(
        space,
        plans=[
            [(2.0, 4), (4.0, 2)],
            [(4.0, 2), (2.0, 4)],
        ],
        top_n=2,
    )
    for r in reports:
        print(r.jsondict)
    assert 2 == len(reports)


def test_asha(tmpdir):
    TUNE_OBJECT_FACTORY.set_temp_path(str(tmpdir))

    space = keras_space(MockSpec, l1=RandInt(8, 16), l2=RandInt(8, 24))
    with raises(TuneCompileError):
        suggest_keras_models_by_continuous_asha(space, plan=[(2.0, 4), (4.0, 2)])

    space = space.sample(6, 0)
    reports = suggest_keras_models_by_continuous_asha(
        space, plan=[(2.0, 4), (4.0, 2)], top_n=2
    )
    for r in reports:
        print(r.jsondict)
    assert 2 == len(reports)



import json

import numpy as np
from tune._utils import normalize_hp


def test_normalize_hp():
    assert isinstance(np.int64(10), np.int64)
    assert 10 == normalize_hp(np.int64(10))
    assert not isinstance(normalize_hp(np.int64(10)), np.int64)

    assert json.dumps(dict(a=[0, 1], b=1.1, c="x")) == json.dumps(
        normalize_hp(dict(a=[np.int64(0), 1], b=np.float64(1.1), c="x"))
    )



import tempfile
from typing import Any, Dict, List, Tuple

from fs.base import FS as FSBase
from tensorflow import keras
from triad import FileSystem


class KerasTrainingSpec:
    def __init__(self, params: Dict[str, Any], dfs: Dict[str, Any]):
        self._params = params
        self._dfs = dfs

    @property
    def params(self) -> Dict[str, Any]:
        return self._params

    @property
    def dfs(self) -> Dict[str, Any]:
        return self._dfs

    def finalize(self) -> None:
        pass

    def generate_sort_metric(self, metric: float) -> float:
        raise NotImplementedError  # pragma: no cover

    def get_fit_metric(self, history: keras.callbacks.History) -> float:
        raise NotImplementedError  # pragma: no cover

    def get_fit_params(self) -> Tuple[List[Any], Dict[str, Any]]:
        raise NotImplementedError  # pragma: no cover

    def get_compile_params(self) -> Dict[str, Any]:
        raise NotImplementedError  # pragma: no cover

    def get_model(self) -> keras.models.Model:
        raise NotImplementedError  # pragma: no cover

    def save_checkpoint(self, fs: FSBase, model: keras.models.Model) -> None:
        with tempfile.NamedTemporaryFile(suffix=".h5") as tf:
            model.save_weights(tf.name)
            with open(tf.name, "rb") as fin:
                fs.writefile("model.h5", fin)

    def load_checkpoint(self, fs: FSBase, model: keras.models.Model) -> None:
        with tempfile.NamedTemporaryFile(suffix=".h5") as tf:
            local_fs = FileSystem()
            with fs.open("model.h5", "rb") as fin:
                local_fs.writefile(tf.name, fin)
            model.load_weights(tf.name)

    def compile_model(self, **add_kwargs: Any) -> keras.models.Model:
        params = dict(self.get_compile_params())
        params.update(add_kwargs)
        model = self.get_model()
        model.compile(**params)
        return model

    def fit(self, **add_kwargs: Any) -> keras.callbacks.History:
        args, kwargs = self.get_fit_params()
        kwargs = dict(kwargs)
        kwargs.update(add_kwargs)
        model = self.compile_model()
        metric = model.fit(*args, **kwargs)
        self.finalize()
        return metric

    def compute_sort_metric(self, **add_kwargs: Any) -> float:
        metric = self.get_fit_metric(self.fit(**add_kwargs))
        return self.generate_sort_metric(metric)



from typing import Any, Dict, List, Optional

from tune.concepts.space.parameters import decode_params, encode_params


class Trial:
    def __init__(
        self,
        trial_id: str,
        params: Dict[str, Any],
        metadata: Optional[Dict[str, Any]] = None,
        keys: Optional[List[str]] = None,
        dfs: Optional[Dict[str, Any]] = None,
        raw: bool = False,
    ):
        self._trial_id = trial_id
        self._params = params if raw else decode_params(params)
        self._metadata = metadata or {}
        self._keys = keys or []
        self._dfs = dfs or {}

    def copy(self) -> "Trial":
        return Trial(
            trial_id=self._trial_id,
            params=self._params,
            metadata=self._metadata,
            keys=self._keys,
            dfs=self._dfs,
            raw=True,
        )

    def __copy__(self) -> "Trial":
        return self.copy()

    def __deepcopy__(self, memo: Any) -> "Trial":
        return self.copy()

    @property
    def trial_id(self) -> str:
        return self._trial_id

    @property
    def params(self) -> Dict[str, Any]:
        return self._params

    @property
    def keys(self) -> List[str]:
        return self._keys

    @property
    def dfs(self) -> Dict[str, Any]:
        return self._dfs

    def with_dfs(self, dfs: Dict[str, Any]) -> "Trial":
        if len(dfs) == 0 and len(self.dfs) == 0:
            return self
        t = self.copy()
        t._dfs = dfs
        return t

    def with_params(self, params: Dict[str, Any], raw: bool = False) -> "Trial":
        t = self.copy()
        t._params = params if raw else decode_params(params)
        return t

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @property
    def jsondict(self) -> Dict[str, Any]:
        return {
            "trial_id": self.trial_id,
            "params": encode_params(self.params),
            "metadata": self.metadata,
            "keys": self.keys,
        }

    @staticmethod
    def from_jsondict(data: Dict[str, Any]) -> "Trial":
        return Trial(**data)



# flake8: noqa

from tune._utils.collections import _EMPTY_ITER, dict_product, product, safe_iter
from tune._utils.execution import run_monitored_process
from tune._utils.math import (
    adjust_high,
    normal_to_continuous,
    normal_to_discrete,
    normal_to_integers,
    uniform_to_choice,
    uniform_to_continuous,
    uniform_to_discrete,
    uniform_to_integers,
)
from tune._utils.test import assert_close
from tune._utils.values import normalize_hp



from typing import Any, Callable, Dict, Optional

from tune.concepts.flow.report import TrialReport
from tune.concepts.flow.trial import Trial


class TrialDecision:
    def __init__(
        self,
        report: TrialReport,
        budget: float,
        should_checkpoint: bool,
        reason: str = "",
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self._report = report
        self._budget = budget
        self._should_checkpoint = should_checkpoint
        self._reason = reason
        self._metadata = metadata or {}

    def __copy__(self) -> "TrialDecision":
        return self

    def __deepcopy__(self, memo: Any) -> "TrialDecision":
        return self

    @property
    def report(self) -> TrialReport:
        return self._report

    @property
    def trial(self) -> Trial:
        return self.report.trial

    @property
    def trial_id(self) -> str:
        return self.trial.trial_id

    @property
    def budget(self) -> float:
        return self._budget

    @property
    def should_stop(self) -> bool:
        return self.budget <= 0

    @property
    def should_checkpoint(self) -> bool:
        return self._should_checkpoint

    @property
    def reason(self) -> str:
        return self._reason

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @property
    def jsondict(self) -> Dict[str, Any]:
        return {
            "report": self.report.jsondict,
            "budget": self.budget,
            "should_checkpoint": self.should_checkpoint,
            "reason": self.reason,
            "metadata": self.metadata,
        }

    @staticmethod
    def from_jsondict(data: Dict[str, Any]) -> "TrialDecision":
        report = TrialReport.from_jsondict(data.pop("report"))
        return TrialDecision(report=report, **data)


class TrialJudge(object):
    def __init__(self, monitor: Optional["Monitor"] = None):
        self.reset_monitor(monitor)

    @property
    def monitor(self) -> "Monitor":
        assert self._trial_judge_monitor is not None
        return self._trial_judge_monitor

    def reset_monitor(self, monitor: Optional["Monitor"] = None) -> None:
        self._trial_judge_monitor = monitor or Monitor()

    def can_accept(self, trial: Trial) -> bool:  # pragma: no cover
        raise NotImplementedError

    def get_budget(self, trial: Trial, rung: int) -> float:  # pragma: no cover
        raise NotImplementedError

    def judge(self, report: TrialReport) -> TrialDecision:  # pragma: no cover
        raise NotImplementedError


class RemoteTrialJudge(TrialJudge):
    def __init__(self, entrypoint: Callable[[str, Dict[str, Any]], Any]):
        super().__init__()
        self._entrypoint = entrypoint
        self._report: Optional[TrialReport] = None

    @property
    def report(self) -> Optional[TrialReport]:
        return self._report

    def can_accept(self, trial: Trial) -> bool:
        return self._entrypoint("can_accept", dict(trial=trial.jsondict))

    def judge(self, report: TrialReport) -> TrialDecision:
        self._report = report
        return TrialDecision.from_jsondict(self._entrypoint("judge", report.jsondict))

    def get_budget(self, trial: Trial, rung: int) -> float:
        return self._entrypoint("get_budget", dict(trial=trial.jsondict, rung=rung))


class NoOpTrailJudge(TrialJudge):
    def can_accept(self, trial: Trial) -> bool:  # pragma: no cover
        return True

    def get_budget(self, trial: Trial, rung: int) -> float:  # pragma: no cover
        return 0.0

    def judge(self, report: TrialReport) -> TrialDecision:  # pragma: no cover
        self.monitor.on_report(report)
        return TrialDecision(report, 0.0, False)


class TrialCallback:
    def __init__(self, judge: TrialJudge):
        self._judge = judge

    def entrypoint(self, name, kwargs: Dict[str, Any]) -> Any:
        if name == "can_accept":
            return self._judge.can_accept(Trial.from_jsondict(kwargs["trial"]))
        if name == "judge":
            return self._judge.judge(TrialReport.from_jsondict(kwargs)).jsondict
        if name == "get_budget":
            return self._judge.get_budget(
                Trial.from_jsondict(kwargs["trial"]), kwargs["rung"]
            )
        raise NotImplementedError  # pragma: no cover


class Monitor:
    def __init__(self):
        self._judge: Optional[TrialJudge] = None

    def on_report(self, report: TrialReport) -> None:  # pragma: no cover
        pass

    def on_get_budget(
        self, trial: Trial, rung: int, budget: float
    ) -> None:  # pragma: no cover
        pass

    def on_judge(self, decision: TrialDecision) -> None:  # pragma: no cover
        pass

    def initialize(self) -> None:  # pragma: no cover
        pass

    def finalize(self) -> None:  # pragma: no cover
        pass



from copy import deepcopy
from typing import Any, Dict, Iterable, List, Tuple, no_type_check

import numpy as np
from triad import assert_or_throw
from tune._utils import dict_product, product
from tune.concepts.space.parameters import Grid, StochasticExpression, encode_params


class Space(object):
    def __init__(self, **kwargs: Any):
        self._value = deepcopy(kwargs)
        self._grid: List[Tuple[Any, Any, Grid]] = []
        self._rand: List[Tuple[Any, Any, StochasticExpression]] = []
        for k in self._value.keys():
            self._search(self._value, k)

    def __iter__(self) -> Iterable[Dict[str, Any]]:
        if len(self._grid) == 0:
            yield deepcopy(self._value)
            return
        rv = [[(x, y, z, v) for v in z] for x, y, z in self._grid]  # type: ignore
        for tps in product(rv, safe=True, remove_empty=True):  # type: ignore
            # overwrite Grid with one value
            for parent, key, _, v in tps:
                parent[key] = v
            res = deepcopy(self._value)
            # undo the overwrite
            for parent, key, orig, _ in tps:
                parent[key] = orig
            yield res

    @property
    def has_random_parameter(self):
        return len(self._rand) > 0

    def sample(self, n: int, seed: Any = None) -> "Space":
        if n <= 0 or not self.has_random_parameter:
            return self
        if seed is not None:
            np.random.seed(seed)
        return VerticalSpace(*self._sample_to_spaces(n))

    def _sample_to_spaces(self, n: int) -> List["Space"]:
        spaces: List["Space"] = []
        rv = [(x, y, z, z.generate_many(n)) for x, y, z in self._rand]
        for i in range(n):
            # overwrite StochasticExpressions with random values
            for parent, key, _, values in rv:
                parent[key] = values[i]
            space = Space(**self._value)
            # undo the overwrite
            for parent, key, orig, _ in rv:
                parent[key] = orig
            spaces.append(space)
        return spaces

    def encode(self) -> Iterable[Any]:
        for s in self:  # type: ignore
            yield encode_params(s)

    def __mul__(self, other: Any) -> "HorizontalSpace":
        return HorizontalSpace(self, other)

    def __add__(self, other: Any) -> "VerticalSpace":
        return VerticalSpace(self, other)

    def __radd__(self, other: Any) -> "Space":
        assert_or_throw(
            other is None or (isinstance(other, int) and other == 0), ValueError(other)
        )
        return self

    def _search(self, parent: Any, key: Any) -> None:
        node = parent[key]
        if isinstance(node, Grid):
            self._grid.append((parent, key, node))
        elif isinstance(node, StochasticExpression):
            self._rand.append((parent, key, node))
        elif isinstance(node, dict):
            for k in node.keys():
                self._search(node, k)
        elif isinstance(node, list):
            for i in range(len(node)):
                self._search(node, i)


class HorizontalSpace(Space):
    def __init__(self, *args: Any, **kwargs: Any):
        self._spaces: List[VerticalSpace] = []
        for x in args:
            if isinstance(x, HorizontalSpace):
                self._spaces.append(VerticalSpace(x))
            elif isinstance(x, VerticalSpace):
                self._spaces.append(x)
            elif isinstance(x, Space):
                self._spaces.append(VerticalSpace(x))
            elif isinstance(x, dict):
                self._spaces.append(VerticalSpace(HorizontalSpace(**x)))
            elif isinstance(x, list):
                self._spaces.append(VerticalSpace(*x))
            else:
                raise ValueError(f"{x} is invalid")
        self._dict = {k: _SpaceValue(v) for k, v in kwargs.items()}

    @no_type_check  # TODO: remove this?
    def __iter__(self) -> Iterable[Dict[str, Any]]:
        dicts = list(dict_product(self._dict, safe=True))
        for spaces in product(
            [g.spaces for g in self._spaces], safe=True, remove_empty=True
        ):
            for comb in product(list(spaces) + [dicts], safe=True, remove_empty=True):
                res: Dict[str, Any] = {}
                for d in comb:
                    res.update(d)
                yield res

    @property
    def has_random_parameter(self):
        return any(x.has_random_parameter for x in self._spaces)

    def _sample_to_spaces(self, n: int) -> List[Space]:
        lists = [s._sample_to_spaces(n) for s in self._spaces]
        return [HorizontalSpace(*args) for args in zip(*lists)]


class VerticalSpace(Space):
    def __init__(self, *args: Any):
        self._spaces: List[Space] = []
        for x in args:
            if isinstance(x, Space):
                self._spaces.append(x)
            elif isinstance(x, dict):
                self._spaces.append(Space(**x))
            elif isinstance(x, list):
                self._spaces.append(VerticalSpace(*x))
            else:
                raise ValueError(f"{x} is invalid")

    @property
    def spaces(self) -> List[Space]:
        return self._spaces

    def __iter__(self) -> Iterable[Dict[str, Any]]:
        for space in self._spaces:
            yield from space  # type: ignore

    @property
    def has_random_parameter(self):
        return any(x.has_random_parameter for x in self._spaces)

    def _sample_to_spaces(self, n: int) -> List[Space]:
        lists = [s._sample_to_spaces(n) for s in self._spaces]
        return [VerticalSpace(*args) for args in zip(*lists)]


class _SpaceValue(object):
    def __init__(self, value: Any):
        self.value = value

    @no_type_check  # TODO: remove this?
    def __iter__(self) -> Iterable[Any]:
        if isinstance(self.value, (HorizontalSpace, VerticalSpace)):
            yield from self.value
        elif isinstance(self.value, dict):
            yield from dict_product(
                {k: _SpaceValue(v) for k, v in self.value.items()}, safe=True
            )
        elif isinstance(self.value, list):
            yield from product([_SpaceValue(v) for v in self.value], safe=True)
        else:
            yield self.value



from pytest import raises
from tune import Grid
from tune.constants import SPACE_MODEL_NAME

from tune_sklearn.utils import sk_space, to_sk_model, to_sk_model_expr


def test_conversion():
    model = to_sk_model("sklearn.linear_model.LinearRegression")
    expr = to_sk_model_expr(model)
    assert "sklearn.linear_model._base.LinearRegression" == expr
    expr = to_sk_model_expr("sklearn.linear_model.LinearRegression")
    assert "sklearn.linear_model._base.LinearRegression" == expr

    raises(TypeError, lambda: to_sk_model("int"))


def test_sk_space():
    rows = list(
        sk_space(
            "sklearn.linear_model.LinearRegression", fit_intercept=Grid(True, False)
        )
    )
    assert 2 == len(rows)
    assert "sklearn.linear_model._base.LinearRegression" == rows[0][SPACE_MODEL_NAME]



import os
import pickle
from tune.api.factory import TUNE_OBJECT_FACTORY
from typing import Any, Optional, Tuple
from uuid import uuid4

import numpy as np
import pandas as pd
from sklearn.metrics import get_scorer
from sklearn.model_selection import cross_val_score
from triad import FileSystem
from tune import NonIterativeObjectiveFunc, Trial, TrialReport
from tune.constants import (
    SPACE_MODEL_NAME,
    TUNE_DATASET_DF_DEFAULT_NAME,
    TUNE_DATASET_VALIDATION_DF_DEFAULT_NAME,
)

from tune_sklearn.utils import to_sk_model, to_sk_model_expr


class SKObjective(NonIterativeObjectiveFunc):
    def __init__(
        self,
        scoring: Any,
        feature_prefix: str = "",
        label_col: str = "label",
        checkpoint_path: Optional[str] = None,
    ) -> None:
        super().__init__()
        self._last_id = ""
        self._model_type: Any = None
        self._model_expr: str = ""

        self._scoring = scoring
        self._feature_prefix = feature_prefix
        self._label_col = label_col
        if checkpoint_path is None:
            self._checkpoint_path = checkpoint_path
        else:
            self._checkpoint_path = TUNE_OBJECT_FACTORY.get_path_or_temp(
                checkpoint_path
            )

    def generate_sort_metric(self, value: float) -> float:
        return -value

    def run(self, trial: Trial) -> TrialReport:
        params = dict(trial.params)
        if trial.trial_id != self._last_id:
            self._model_type = to_sk_model(params.pop(SPACE_MODEL_NAME))
            self._model_expr = to_sk_model_expr(self._model_type)
            self._train_x, self._train_y = self._reset_xy(
                trial.dfs[TUNE_DATASET_DF_DEFAULT_NAME]
            )
            self._test_x, self._test_y = self._reset_xy(
                trial.dfs[TUNE_DATASET_VALIDATION_DF_DEFAULT_NAME]
            )
            self._last_id = trial.trial_id
        else:
            params.pop(SPACE_MODEL_NAME)

        model = self._model_type(**params).fit(self._train_x, self._train_y)
        metric = get_scorer(self._scoring)(model, self._test_x, self._test_y)
        metadata = dict(model=self._model_expr)
        if self._checkpoint_path is not None:
            fp = os.path.join(self._checkpoint_path, str(uuid4()) + ".pkl")
            with FileSystem().openbin(fp, mode="wb") as f:
                pickle.dump(model, f)
            metadata["checkpoint_path"] = fp
        return TrialReport(
            trial,
            metric=metric,
            metadata=metadata,
            sort_metric=self.generate_sort_metric(metric),
        )

    def _reset_xy(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        train_df = df.sample(frac=1, random_state=0).reset_index(drop=True)

        train_x = train_df.drop([self._label_col], axis=1)
        cols = [x for x in train_x.columns if x.startswith(self._feature_prefix)]
        return train_x[cols], train_df[self._label_col]


class SKCVObjective(SKObjective):
    def __init__(
        self,
        scoring: Any,
        cv: int = 5,
        feature_prefix: str = "",
        label_col: str = "label",
        checkpoint_path: Optional[str] = None,
    ) -> None:
        super().__init__(
            scoring=scoring,
            feature_prefix=feature_prefix,
            label_col=label_col,
            checkpoint_path=checkpoint_path,
        )
        self._cv = cv

    def run(self, trial: Trial) -> TrialReport:
        params = dict(trial.params)
        if trial.trial_id != self._last_id:
            self._model_type = to_sk_model(params.pop(SPACE_MODEL_NAME))
            self._model_expr = to_sk_model_expr(self._model_type)
            self._train_x, self._train_y = self._reset_xy(
                trial.dfs[TUNE_DATASET_DF_DEFAULT_NAME]
            )
            self._last_id = trial.trial_id
        else:
            params.pop(SPACE_MODEL_NAME)

        model = self._model_type(**params)
        s = cross_val_score(
            model, self._train_x, self._train_y, cv=self._cv, scoring=self._scoring
        )
        metadata = dict(model=self._model_expr, cv_scores=[float(x) for x in s])
        if self._checkpoint_path is not None:
            model.fit(self._train_x, self._train_y)
            fp = os.path.join(self._checkpoint_path, str(uuid4()) + ".pkl")
            with FileSystem().openbin(fp, mode="wb") as f:
                pickle.dump(model, f)
            metadata["checkpoint_path"] = fp
        metric = float(np.mean(s))
        return TrialReport(
            trial,
            metric=metric,
            metadata=metadata,
            sort_metric=self.generate_sort_metric(metric),
        )



from tune.constants import TUNE_STOPPER_DEFAULT_CHECK_INTERVAL
from typing import Any, Callable, Optional

from tune._utils import run_monitored_process
from tune.concepts.flow import Trial, TrialReport


class NonIterativeObjectiveFunc:
    def generate_sort_metric(self, value: float) -> float:  # pragma: no cover
        return value

    def run(self, trial: Trial) -> TrialReport:  # pragma: no cover
        raise NotImplementedError


class NonIterativeObjectiveLocalOptimizer:
    @property
    def distributable(self) -> bool:
        return True

    def run(self, func: NonIterativeObjectiveFunc, trial: Trial) -> TrialReport:
        # TODO: how to utilize execution_engine?
        return func.run(trial)

    def run_monitored_process(
        self,
        func: NonIterativeObjectiveFunc,
        trial: Trial,
        stop_checker: Callable[[], bool],
        interval: Any = TUNE_STOPPER_DEFAULT_CHECK_INTERVAL,
    ) -> TrialReport:
        return run_monitored_process(
            self.run, [func, trial], {}, stop_checker=stop_checker, interval=interval
        )


def validate_noniterative_objective(
    func: NonIterativeObjectiveFunc,
    trial: Trial,
    validator: Callable[[TrialReport], None],
    optimizer: Optional[NonIterativeObjectiveLocalOptimizer] = None,
) -> None:
    _optimizer = optimizer or NonIterativeObjectiveLocalOptimizer()
    validator(_optimizer.run_monitored_process(func, trial, lambda: False, "1sec"))



from typing import Any, Callable, Dict, Iterable

from triad import FileSystem
from tune.constants import TUNE_REPORT_ADD_SCHEMA
from tune.concepts.dataset import StudyResult, TuneDataset, get_trials_from_row
from tune.iterative.objective import IterativeObjectiveFunc
from tune.concepts.flow import RemoteTrialJudge, TrialCallback, TrialJudge


class IterativeStudy:
    def __init__(self, objective: IterativeObjectiveFunc, checkpoint_path: str):
        self._objective = objective
        self._checkpoint_path = checkpoint_path

    def optimize(self, dataset: TuneDataset, judge: TrialJudge) -> StudyResult:
        callback = TrialCallback(judge)

        res = dataset.data.per_row().transform(
            self._compute,
            schema=f"*,{TUNE_REPORT_ADD_SCHEMA}",
            callback=callback.entrypoint,
        )

        return StudyResult(dataset=dataset, result=res)

    def _compute(
        self,
        df: Iterable[Dict[str, Any]],
        entrypoint: Callable[[str, Dict[str, Any]], Any],
    ) -> Iterable[Dict[str, Any]]:
        ck_fs = FileSystem().makedirs(self._checkpoint_path, recreate=True)
        for row in df:
            for trial in get_trials_from_row(row):
                rjudge = RemoteTrialJudge(entrypoint)
                self._objective.copy().run(trial, rjudge, ck_fs)
                if rjudge.report is not None:
                    yield rjudge.report.fill_dict(dict(row))



from fugue.exceptions import FugueWorkflowCompileError, FugueWorkflowRuntimeError


class TuneCompileError(FugueWorkflowCompileError):
    pass


class TuneRuntimeError(FugueWorkflowRuntimeError):
    pass


class TuneInterrupted(TuneRuntimeError):
    pass



from typing import Any, Dict, Iterable
import numpy as np
from fugue.workflow.workflow import FugueWorkflow
from tune import Space, Trial, TrialDecision, TrialReport
from tune.constants import TUNE_REPORT_METRIC
from tune.concepts.dataset import TuneDatasetBuilder
from tune.iterative.objective import IterativeObjectiveFunc
from tune.iterative.study import IterativeStudy
from tune.concepts.flow import TrialJudge


def f(x, a, b):
    return -np.log(x + 0.01) * a + b


class F(IterativeObjectiveFunc):
    def __init__(self) -> None:
        self.step = 0
        super().__init__()

    def copy(self) -> "F":
        return F()

    def initialize(self) -> None:
        assert self.step == 0  # because of copy

    def run_single_iteration(self) -> TrialReport:
        self.step += 1
        trial = self.current_trial
        return TrialReport(
            trial=trial, metric=f(self.step, trial.params["a"], trial.params["b"])
        )


class J(TrialJudge):
    def __init__(self, schedule):
        super().__init__()
        self.schedule = schedule

    def can_accept(self, trial: Trial) -> bool:
        return True

    def get_budget(self, trial: Trial, rung: int) -> float:
        return float(self.schedule[rung]) if rung < len(self.schedule) else 0.0

    def judge(self, report: TrialReport) -> TrialDecision:
        return TrialDecision(
            report,
            budget=self.get_budget(report.trial, report.rung + 1),
            should_checkpoint=False,
            metadata={},
        )


def test_iterative_study(tmpdir):
    def assert_metric(df: Iterable[Dict[str, Any]], metric: float) -> None:
        for row in df:
            assert row[TUNE_REPORT_METRIC] < metric

    study = IterativeStudy(F(), str(tmpdir))
    space = sum(
        Space(a=a, b=b)
        for a, b in [(1.1, 0.2), (0.8, -0.2), (1.2, -0.1), (0.7, 0.3), (1.0, 1.5)]
    )
    dag = FugueWorkflow()
    dataset = TuneDatasetBuilder(space, str(tmpdir)).build(dag)
    result = study.optimize(
        dataset,
        J([1, 2, 3, 4]),
    )
    result.result(1).show()
    result.result(1).output(assert_metric, params=dict(metric=-2.8))

    dag.run()



from tests.tune_tensorflow.mock import MockSpec
from tune_tensorflow.utils import (
    extract_keras_spec,
    to_keras_spec,
    to_keras_spec_expr,
    keras_space,
)


def test_keras_space():
    space = keras_space(MockSpec, a=1, b=2)
    spec = extract_keras_spec(list(space.encode())[0], {})
    assert spec == MockSpec
    spec = extract_keras_spec(
        list(space.encode())[0], {to_keras_spec_expr(MockSpec): "dummy"}
    )
    assert "dummy" == spec


def test_to_keras_spec():
    expr = to_keras_spec_expr(MockSpec)
    assert to_keras_spec(expr) == MockSpec

    expr = to_keras_spec_expr(to_keras_spec_expr(MockSpec))
    assert to_keras_spec(expr) == MockSpec


from typing import Any, List, Optional, Tuple

from tune import (
    Space,
    TrialReport,
    suggest_by_continuous_asha,
    suggest_by_hyperband,
    suggest_by_sha,
)

from tune_tensorflow.objective import KerasObjective
from tune_tensorflow.utils import _TYPE_DICT


def suggest_keras_models_by_sha(
    space: Space,
    plan: List[Tuple[float, int]],
    train_df: Any = None,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    monitor: Any = None,
    distributed: Optional[bool] = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    return suggest_by_sha(
        objective=KerasObjective(_TYPE_DICT),
        space=space,
        plan=plan,
        train_df=train_df,
        temp_path=temp_path,
        partition_keys=partition_keys,
        top_n=top_n,
        monitor=monitor,
        distributed=distributed,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )


def suggest_keras_models_by_hyperband(
    space: Space,
    plans: List[List[Tuple[float, int]]],
    train_df: Any = None,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    monitor: Any = None,
    distributed: Optional[bool] = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    return suggest_by_hyperband(
        objective=KerasObjective(_TYPE_DICT),
        space=space,
        plans=plans,
        train_df=train_df,
        temp_path=temp_path,
        partition_keys=partition_keys,
        top_n=top_n,
        monitor=monitor,
        distributed=distributed,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )


def suggest_keras_models_by_continuous_asha(
    space: Space,
    plan: List[Tuple[float, int]],
    train_df: Any = None,
    temp_path: str = "",
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    monitor: Any = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    return suggest_by_continuous_asha(
        objective=KerasObjective(_TYPE_DICT),
        space=space,
        plan=plan,
        train_df=train_df,
        temp_path=temp_path,
        partition_keys=partition_keys,
        top_n=top_n,
        monitor=monitor,
        execution_engine=execution_engine,
        execution_engine_conf=execution_engine_conf,
    )



from tensorflow.keras import layers, models
from tensorflow.keras.datasets import boston_housing

from tune_tensorflow import KerasTrainingSpec


class MockSpec(KerasTrainingSpec):
    def __init__(self, params, dfs):
        super().__init__(params, dfs)
        (self.train_data, self.train_targets), (
            self.test_data,
            self.test_targets,
        ) = boston_housing.load_data()

    def get_model(self):
        l1, l2 = self.params["l1"], self.params["l2"]
        model = models.Sequential()
        model.add(
            layers.Dense(l1, activation="relu", input_shape=(self.train_data.shape[1],))
        )
        model.add(layers.Dense(l2, activation="relu"))
        model.add(layers.Dense(1))
        return model

    def get_compile_params(self):
        return dict(optimizer="rmsprop", loss="mse", metrics=["mae"])

    def get_fit_params(self):
        return [self.train_data, self.train_targets], dict(
            validation_data=(self.test_data, self.test_targets), shuffle=True
        )

    def get_fit_metric(self, history):
        return float(history.history["val_mae"][-1])

    def generate_sort_metric(self, metric: float) -> float:
        return metric



from typing import Any

from tune.noniterative.objective import NonIterativeObjectiveLocalOptimizer
from tune_test.local_optmizer import NonIterativeObjectiveLocalOptimizerTests

from tune_optuna.optimizer import OptunaLocalOptimizer


class HyperoptLocalOptimizerTests(NonIterativeObjectiveLocalOptimizerTests.Tests):
    def make_optimizer(self, **kwargs: Any) -> NonIterativeObjectiveLocalOptimizer:
        return OptunaLocalOptimizer(**kwargs)



from tune._utils.math import adjust_high
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np
from hyperopt import STATUS_OK, Trials, fmin, hp, tpe
from tune.noniterative.objective import (
    NonIterativeObjectiveFunc,
    NonIterativeObjectiveLocalOptimizer,
)
from tune.concepts.space.parameters import Choice, Rand, RandInt, StochasticExpression
from tune.concepts.flow import Trial, TrialReport


class HyperoptLocalOptimizer(NonIterativeObjectiveLocalOptimizer):
    def __init__(
        self,
        max_iter: int,
        seed: int = 0,
        kwargs_func: Optional[
            Callable[[NonIterativeObjectiveFunc, Trial], Dict[str, Any]]
        ] = None,
    ):
        self._max_iter = max_iter
        self._seed = seed
        self._kwargs_func = kwargs_func

    def run(self, func: NonIterativeObjectiveFunc, trial: Trial) -> TrialReport:
        static_params, stochastic_params, postprocess = self._split(trial.params)
        stochastic_keys = list(stochastic_params.keys())
        if len(stochastic_keys) == 0:
            return func.run(trial)

        def obj(args) -> Dict[str, Any]:
            params = {k: postprocess[k](v) for k, v in zip(stochastic_keys, args)}
            params.update(static_params)
            report = func.run(trial.with_params(params))
            return {
                "loss": report.sort_metric,
                "status": STATUS_OK,
                "report": report,
            }

        trials = Trials()
        kwargs: Dict[str, Any] = dict(
            algo=tpe.suggest,
            max_evals=self._max_iter,
            trials=trials,
            show_progressbar=False,
            rstate=np.random.RandomState(self._seed),
        )
        if self._kwargs_func is not None:
            kwargs.update(self._kwargs_func(func, trial))

        fmin(obj, space=list(stochastic_params.values()), **kwargs)
        return trials.best_trial["result"]["report"]

    def _split(
        self, kwargs: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        static_params: Dict[str, Any] = {}
        stochastic_params: Dict[str, Any] = {}
        postprocess: Dict[str, Any] = {}
        for k, v in kwargs.items():
            if isinstance(v, StochasticExpression):
                if isinstance(v, RandInt):
                    stochastic_params[k], postprocess[k] = _convert_randint(k, v)
                elif isinstance(v, Rand):
                    stochastic_params[k], postprocess[k] = _convert_rand(k, v)
                elif isinstance(v, Choice):
                    stochastic_params[k], postprocess[k] = _convert_choice(k, v)
                else:
                    # TODO: normal rand and normal randint
                    raise NotImplementedError(v)  # pragma: no cover
            else:
                static_params[k] = v
        return static_params, stochastic_params, postprocess


def _convert_randint(k: str, v: RandInt) -> Any:
    _high = adjust_high(
        0, v.high - v.low, v.q, include_high=v.include_high  # type:ignore
    )
    n = int(np.round(_high / v.q))
    if not v.log:
        return hp.randint(k, 0, n) * v.q + v.low, lambda x: int(np.round(x))
    return (
        hp.qloguniform(k, np.log(v.q), np.log(_high), q=v.q) + v.low - v.q,
        lambda x: int(np.round(x)),
    )


def _convert_rand(k: str, v: Rand) -> Any:
    if v.q is None:
        if not v.log:
            return hp.uniform(k, v.low, v.high), lambda x: x
        else:
            return hp.loguniform(k, np.log(v.low), np.log(v.high)), lambda x: x
    else:
        _high = adjust_high(0, v.high - v.low, v.q, include_high=v.include_high)
        n = int(np.round(_high / v.q))
        if not v.log:
            return hp.randint(k, 0, n) * v.q + v.low, lambda x: x
        else:
            _high = max(0, _high - v.q)
            return hp.qloguniform(k, 0, np.log(_high), q=v.q) + v.low, lambda x: x


def _convert_choice(k: str, v: Choice) -> Any:
    return hp.choice(k, v.values), lambda x: x



from typing import Any, Callable, List, Optional

from fugue import FugueWorkflow
from triad import assert_or_throw
from tune.concepts.dataset import TuneDataset, TuneDatasetBuilder
from tune.concepts.flow import Monitor
from tune.concepts.space import Space
from tune.constants import (
    TUNE_DATASET_DF_DEFAULT_NAME,
    TUNE_DATASET_VALIDATION_DF_DEFAULT_NAME,
)
from tune.exceptions import TuneCompileError
from tune.iterative.objective import IterativeObjectiveFunc
from tune.noniterative.convert import to_noniterative_objective
from tune.noniterative.objective import (
    NonIterativeObjectiveFunc,
    NonIterativeObjectiveLocalOptimizer,
)
from tune.noniterative.stopper import NonIterativeStopper


class TuneObjectFactory:
    def __init__(self):
        self._iterative_objective_converter = self._object_to_iterative_objective
        self._noniterative_objective_converter = self._object_to_noniterative_objective
        self._noniterative_local_optimizer_converter = (
            self._object_to_noniterative_local_optimizer
        )
        self._monitor_converter = self._object_to_monitor
        self._stopper_converter = self._object_to_stopper
        self._tmp = ""

    def set_iterative_objective_converter(
        self, func: Callable[[Any], IterativeObjectiveFunc]
    ) -> None:
        self._iterative_objective_converter = func

    def set_noniterative_objective_converter(
        self, func: Callable[[Any], NonIterativeObjectiveFunc]
    ) -> None:
        self._noniterative_objective_converter = func

    def set_noniterative_local_optimizer_converter(
        self, func: Callable[[Any], NonIterativeObjectiveLocalOptimizer]
    ) -> None:
        self._noniterative_local_optimizer_converter = func

    def set_noniterative_stopper_converter(
        self, func: Callable[[Any], NonIterativeStopper]
    ) -> None:
        self._stopper_converter = func

    def set_monitor_converter(self, func: Callable[[Any], Optional[Monitor]]) -> None:
        self._monitor_converter = func

    def set_temp_path(self, path: str) -> None:
        self._tmp = path

    def make_iterative_objective(self, obj: Any) -> IterativeObjectiveFunc:
        return self._iterative_objective_converter(obj)

    def make_noniterative_objective(self, obj: Any) -> NonIterativeObjectiveFunc:
        return self._noniterative_objective_converter(obj)

    def make_noniterative_local_optimizer(
        self, obj: Any
    ) -> NonIterativeObjectiveLocalOptimizer:
        return self._noniterative_local_optimizer_converter(obj)

    def make_monitor(self, obj: Any) -> Optional[Monitor]:
        return self._monitor_converter(obj)

    def make_noniterative_stopper(self, obj: Any) -> Optional[NonIterativeStopper]:
        return self._stopper_converter(obj)

    def make_dataset(
        self,
        dag: FugueWorkflow,
        dataset: Any,
        df: Any = None,
        df_name: str = TUNE_DATASET_DF_DEFAULT_NAME,
        test_df: Any = None,
        test_df_name: str = TUNE_DATASET_VALIDATION_DF_DEFAULT_NAME,
        partition_keys: Optional[List[str]] = None,
        shuffle: bool = True,
        temp_path: str = "",
    ) -> TuneDataset:
        assert_or_throw(dataset is not None, TuneCompileError("dataset can't be None"))
        if isinstance(dataset, TuneDataset):
            assert_or_throw(
                df is None, TuneCompileError("can't set df when dataset is TuneDataset")
            )
            return dataset
        if isinstance(dataset, Space):
            path = self.get_path_or_temp(temp_path)
            builder = TuneDatasetBuilder(dataset, path)
            if df is not None:
                wdf = dag.df(df)
                if partition_keys is not None and len(partition_keys) > 0:
                    wdf = wdf.partition_by(*partition_keys)
                builder.add_df(df_name, wdf)
            if test_df is not None:
                wdf = dag.df(test_df)
                how = "cross"
                if partition_keys is not None and len(partition_keys) > 0:
                    wdf = wdf.partition_by(*partition_keys)
                    how = "inner"
                builder.add_df(test_df_name, wdf, how=how)
            return builder.build(dag, batch_size=1, shuffle=shuffle)
        raise TuneCompileError(f"{dataset} can't be converted to TuneDataset")

    def get_path_or_temp(self, path: str) -> str:
        if path is None or path == "":
            path = self._tmp
        assert_or_throw(path != "", TuneCompileError("path or temp path must be set"))
        return path

    def _object_to_iterative_objective(self, obj: Any) -> IterativeObjectiveFunc:
        assert_or_throw(obj is not None, TuneCompileError("objective can't be None"))
        if isinstance(obj, IterativeObjectiveFunc):
            return obj
        raise TuneCompileError(
            f"{obj} can't be converted to iterative objective function"
        )

    def _object_to_noniterative_objective(self, obj: Any) -> NonIterativeObjectiveFunc:
        assert_or_throw(obj is not None, TuneCompileError("objective can't be None"))
        if isinstance(obj, NonIterativeObjectiveFunc):
            return obj
        if callable(obj):
            return to_noniterative_objective(obj)
        raise TuneCompileError(
            f"{obj} can't be converted to non iterative objective function"
        )

    def _object_to_noniterative_local_optimizer(
        self, obj: Any
    ) -> NonIterativeObjectiveLocalOptimizer:
        if isinstance(obj, NonIterativeObjectiveLocalOptimizer):
            return obj
        if obj is None:
            return NonIterativeObjectiveLocalOptimizer()
        raise TuneCompileError(
            f"{obj} can't be converted to non iterative objective optimizer"
        )

    def _object_to_monitor(self, obj: Any) -> Optional[Monitor]:
        if isinstance(obj, Monitor):
            return obj
        if obj is None:
            return None
        raise TuneCompileError(f"{obj} can't be converted to Monitor")

    def _object_to_stopper(self, obj: Any) -> Optional[NonIterativeStopper]:
        if isinstance(obj, NonIterativeStopper):
            return obj
        if obj is None:
            return None
        raise TuneCompileError(f"{obj} can't be converted to NonIterativeStopper")


TUNE_OBJECT_FACTORY = TuneObjectFactory()



from tune.concepts.dataset import TuneDataset
from fugue.workflow.workflow import FugueWorkflow
from pytest import raises
from tune.api.factory import TuneObjectFactory
from tune.concepts.flow.judge import Monitor
from tune.exceptions import TuneCompileError
from tune.iterative.objective import IterativeObjectiveFunc
from tune.noniterative.convert import to_noniterative_objective
from tune.noniterative.objective import (
    NonIterativeObjectiveFunc,
    NonIterativeObjectiveLocalOptimizer,
)
from tune.noniterative.stopper import NonIterativeStopper

from tune_optuna.optimizer import OptunaLocalOptimizer
import pandas as pd
from tune import Space


def _nobjective(a: int) -> float:
    return 0.0


def test_dataset(tmpdir):
    factory = TuneObjectFactory()
    factory.set_temp_path(str(tmpdir))
    df = pd.DataFrame([[0]], columns=["a"])
    dag = FugueWorkflow()
    data = factory.make_dataset(dag, Space(a=1, b=1), df=df)
    assert isinstance(data, TuneDataset)
    assert factory.make_dataset(dag, data) is data
    # TODO: the unit test is not complete, but it's covered by other functions
    with raises(TuneCompileError):
        factory.make_dataset(dag, 1)


def test_temp_path():
    factory = TuneObjectFactory()
    with raises(TuneCompileError):
        factory.get_path_or_temp("")  # must set temp path
    assert "/x" == factory.get_path_or_temp("/x")

    factory.set_temp_path("/tmp")
    assert "/x" == factory.get_path_or_temp("/x")
    assert "/tmp" == factory.get_path_or_temp("")


def test_noniterative_objective():
    factory = TuneObjectFactory()
    assert isinstance(
        factory.make_noniterative_objective(_nobjective),
        NonIterativeObjectiveFunc,
    )
    with raises(TuneCompileError):
        factory.make_noniterative_objective("x")

    factory.set_noniterative_objective_converter(
        lambda x: to_noniterative_objective(_nobjective)
    )
    assert isinstance(
        factory.make_noniterative_objective("x"),
        NonIterativeObjectiveFunc,
    )


def test_monitor():
    factory = TuneObjectFactory()

    class Mock(Monitor):
        pass

    assert isinstance(
        factory.make_monitor(Mock()),
        Monitor,
    )
    with raises(TuneCompileError):
        factory.make_monitor("x")

    factory.set_monitor_converter(lambda x: Mock())
    assert isinstance(
        factory.make_monitor("x"),
        Monitor,
    )


def test_noniterative_stopper():
    factory = TuneObjectFactory()

    class Mock(NonIterativeStopper):
        pass

    assert isinstance(
        factory.make_noniterative_stopper(Mock()),
        NonIterativeStopper,
    )
    with raises(TuneCompileError):
        factory.make_noniterative_stopper("x")

    factory.set_noniterative_stopper_converter(lambda x: Mock())
    assert isinstance(
        factory.make_noniterative_stopper("x"),
        NonIterativeStopper,
    )


def test_noniterative_local_optimizer_converter():
    factory = TuneObjectFactory()
    assert isinstance(
        factory.make_noniterative_local_optimizer(OptunaLocalOptimizer(0)),
        NonIterativeObjectiveLocalOptimizer,
    )
    with raises(TuneCompileError):
        factory.make_noniterative_local_optimizer("x")

    factory.set_noniterative_local_optimizer_converter(
        lambda x: OptunaLocalOptimizer(0)
    )
    assert isinstance(
        factory.make_noniterative_local_optimizer("x"),
        NonIterativeObjectiveLocalOptimizer,
    )


def test_iterative_objective():
    factory = TuneObjectFactory()

    class Mock(IterativeObjectiveFunc):
        pass

    assert isinstance(
        factory.make_iterative_objective(Mock()),
        IterativeObjectiveFunc,
    )
    with raises(TuneCompileError):
        factory.make_iterative_objective("x")

    factory.set_iterative_objective_converter(lambda x: Mock())
    assert isinstance(
        factory.make_iterative_objective("x"),
        IterativeObjectiveFunc,
    )



from typing import Any

import numpy as np


def normalize_hp(hp: Any) -> Any:
    if isinstance(hp, str):
        return hp
    elif isinstance(hp, list):
        return [normalize_hp(x) for x in hp]
    elif isinstance(hp, dict):
        return {k: normalize_hp(v) for k, v in hp.items()}
    elif isinstance(hp, np.generic):
        return hp.item()
    else:
        return hp



TUNE_TEMP_PATH = "tune.temp.path"
TUNE_PREFIX = "__tune_"
TUNE_DATASET_DF_PREFIX = TUNE_PREFIX + "df__"
TUNE_DATASET_PARAMS_PREFIX = TUNE_PREFIX + "params__"
TUNE_DATASET_TRIALS = TUNE_PREFIX + "trials__"

TUNE_REPORT_ID = TUNE_PREFIX + "trail_id__"
TUNE_REPORT_METRIC = TUNE_PREFIX + "metric__"
TUNE_REPORT = TUNE_PREFIX + "report__"

TUNE_REPORT_ADD_SCHEMA = (
    f"{TUNE_REPORT_ID}:str,{TUNE_REPORT_METRIC}:double,{TUNE_REPORT}:str"
)

TUNE_DATASET_DF_DEFAULT_NAME = TUNE_PREFIX + "_df_"
TUNE_DATASET_VALIDATION_DF_DEFAULT_NAME = TUNE_DATASET_DF_DEFAULT_NAME + "_validation_"

TUNE_STOPPER_DEFAULT_CHECK_INTERVAL = "60sec"

SPACE_PARAM_PREFIX = "__space_"
SPACE_MODEL_NAME = SPACE_PARAM_PREFIX + "_model"



import json

from fugue import (
    ArrayDataFrame,
    DataFrame,
    FugueWorkflow,
    NativeExecutionEngine,
    WorkflowDataFrames,
)
import pickle

import fugue

from tune.constants import (
    TUNE_DATASET_DF_PREFIX,
    TUNE_DATASET_PARAMS_PREFIX,
    TUNE_DATASET_TRIALS,
    TUNE_TEMP_PATH,
)
from tune.concepts.dataset import TuneDatasetBuilder, _to_trail_row, TuneDataset
from tune.concepts.space import Grid, Rand
from tune.concepts.space.spaces import Space
from tune.concepts.flow import Trial


def test_builder(tmpdir):
    space = Space(a=1, b=2, c=Grid(2, 3))
    builder = TuneDatasetBuilder(space, str(tmpdir))

    def assert_count(df: DataFrame, n: int, schema=None) -> None:
        assert len(df.as_array()) == n
        if schema is not None:
            assert df.schema == schema

    # test to_space
    with FugueWorkflow() as dag:
        df = builder.build(dag).data
        df.show()

    df1 = ArrayDataFrame([[0, 1], [1, 1], [0, 2]], "a:int,b:int")

    # test single df
    with FugueWorkflow() as dag:
        builder.add_dfs(WorkflowDataFrames(x=dag.df(df1)))
        dataset = builder.build(dag)
        assert ["x"] == dataset.dfs
        assert [] == dataset.keys
        df = dataset.data
        df.show()
        df.output(
            assert_count,
            params=dict(n=2, schema=f"__tune_df__x:str,{TUNE_DATASET_TRIALS}:str"),
        )

    space = Space(b=Rand(0, 1), a=1, c=Grid(2, 3), d=Grid("a", "b"))
    df2 = ArrayDataFrame([[0, 1], [1, 1], [3, 2]], "a:int,bb:int")
    df3 = ArrayDataFrame([[10, 1], [11, 1], [10, 2]], "a:int,c:int")
    builder = TuneDatasetBuilder(space)
    engine = NativeExecutionEngine(conf={TUNE_TEMP_PATH: str(tmpdir)})

    # test multiple dfs, batch_size and config
    with FugueWorkflow(engine) as dag:
        dfs = WorkflowDataFrames(
            a=dag.df(df1).partition_by("a"), b=dag.df(df2).partition_by("a")
        )
        dataset = (
            builder.add_dfs(dfs, "inner").add_df("c", dag.df(df3), "cross").build(dag)
        )
        assert ["a"] == dataset.keys
        assert ["a", "b", "c"] == dataset.dfs
        df = dataset.data
        df.show()
        df.output(
            assert_count,
            params=dict(
                n=8,
                schema="a:int,__tune_df__a:str,__tune_df__b:str,"
                f"__tune_df__c:str,{TUNE_DATASET_TRIALS}:str",
            ),
        )

        df = builder.build(dag, batch_size=3).data
        df.show()
        df.output(
            assert_count,
            params=dict(
                n=4,
                schema="a:int,__tune_df__a:str,__tune_df__b:str,"
                f"__tune_df__c:str,{TUNE_DATASET_TRIALS}:str",
            ),
        )


def test_dataset(tmpdir):
    space = Space(a=Grid(0, 1, 2, 3, 4), b=Grid(5, 6, 7, 8, 9))
    builder = TuneDatasetBuilder(space, str(tmpdir))

    dag = FugueWorkflow()
    dataset = builder.build(dag)
    ds = dataset.divide([4, 1], 0)
    assert 2 == len(ds)
    ds[0].data.yield_dataframe_as("a")
    ds[1].data.yield_dataframe_as("b")
    res = dag.run()
    assert 25 == len(res["a"].as_array()) + len(res["b"].as_array())
    assert len(res["b"].as_array()) < 10


def test_to_trial_row():
    data1 = {
        "b": 2,
        "a": 1,
        TUNE_DATASET_DF_PREFIX + "x": "x",
        TUNE_DATASET_PARAMS_PREFIX: pickle.dumps(
            [{"b": 10, "a": 11}, {"a": 11, "b": 10}, {"b": 100, "a": 110}],
        ),
    }
    res1 = _to_trail_row(data1, {"m": 1})
    trials1 = [Trial(**p) for p in json.loads(res1[TUNE_DATASET_TRIALS])]
    assert 2 == len(trials1)
    data2 = {
        "a": 1,
        "b": 2,
        TUNE_DATASET_DF_PREFIX + "y": "x",
        TUNE_DATASET_PARAMS_PREFIX: pickle.dumps(
            [{"b": 10, "a": 11}, {"b": 100, "a": 110}],
        ),
    }
    res2 = _to_trail_row(data2, {"m": 1})
    assert TUNE_DATASET_PARAMS_PREFIX not in res2
    trials2 = [Trial(**p) for p in json.loads(res2[TUNE_DATASET_TRIALS])]
    assert 2 == len(trials2)
    assert any(trials2[0].trial_id == x.trial_id for x in trials1)
    assert any(trials2[1].trial_id == x.trial_id for x in trials1)

    data3 = {
        "a": 10,
        "b": 2,
        TUNE_DATASET_DF_PREFIX + "y": "x",
        TUNE_DATASET_PARAMS_PREFIX: pickle.dumps(
            [{"b": 10, "a": 11}, {"b": 100, "a": 110}],
        ),
    }
    res3 = _to_trail_row(data3, {"m": 1})
    trials3 = [Trial(**p) for p in json.loads(res3[TUNE_DATASET_TRIALS])]
    assert 2 == len(trials2)
    assert not any(trials3[0].trial_id == x.trial_id for x in trials1)
    assert not any(trials3[1].trial_id == x.trial_id for x in trials1)



import itertools
from typing import Any, Dict, Iterable, List, Tuple


def dict_product(
    d: Dict[str, Iterable[Any]], safe: bool = True
) -> Iterable[Dict[str, Any]]:
    keys = d.keys()
    arrays = list(d.values())
    if len(arrays) == 0:
        if safe:
            yield {}
        return
    for element in _safe_product(arrays, safe):
        yield {k: v for k, v in zip(keys, element) if v is not _EMPTY_ITER}


def product(
    arrays: List[Iterable[Any]], safe: bool = False, remove_empty: bool = True
) -> Iterable[List[Any]]:
    if len(arrays) == 0:
        if safe:
            yield []
        return
    if remove_empty:
        for x in _safe_product(arrays, safe):
            yield [xx for xx in x if xx is not _EMPTY_ITER]
    else:
        for x in _safe_product(arrays, safe):
            yield [None if xx is _EMPTY_ITER else xx for xx in x]


def safe_iter(it: Iterable[Any], safe: bool = True) -> Iterable[Any]:
    if not safe:
        yield from it
    else:
        n = 0
        for x in it:
            yield x
            n += 1
        if n == 0:
            yield _EMPTY_ITER


def _safe_product(arrays: List[Iterable[Any]], safe: bool = True) -> Iterable[Tuple]:
    if not safe:
        yield from itertools.product(*arrays)
    else:
        arr = [safe_iter(t) for t in arrays]
        yield from itertools.product(*arr)


class _EmptyIter(object):
    pass


_EMPTY_ITER = _EmptyIter()



import json
from typing import Any, List, Optional

from fugue import FugueWorkflow
from tune import (
    TUNE_OBJECT_FACTORY,
    NonIterativeObjectiveLocalOptimizer,
    Space,
    TrialReport,
    optimize_noniterative,
)
from tune.constants import TUNE_REPORT, TUNE_REPORT_METRIC

from tune_sklearn.objective import SKCVObjective, SKObjective


def suggest_sk_models(
    space: Space,
    train_df: Any,
    test_df: Any,
    scoring: str,
    temp_path: str = "",
    feature_prefix: str = "",
    label_col: str = "label",
    save_model: bool = False,
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    local_optimizer: Optional[NonIterativeObjectiveLocalOptimizer] = None,
    monitor: Any = None,
    stopper: Any = None,
    stop_check_interval: Any = None,
    distributed: Optional[bool] = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    dag = FugueWorkflow()
    dataset = TUNE_OBJECT_FACTORY.make_dataset(
        dag,
        space,
        df=train_df,
        test_df=test_df,
        partition_keys=partition_keys,
        temp_path=temp_path,
    )
    objective = SKObjective(
        scoring=scoring,
        feature_prefix=feature_prefix,
        label_col=label_col,
        checkpoint_path=temp_path if save_model else None,
    )
    study = optimize_noniterative(
        objective=objective,
        dataset=dataset,
        optimizer=local_optimizer,
        distributed=distributed,
        monitor=monitor,
        stopper=stopper,
        stop_check_interval=stop_check_interval,
    )
    study.result(top_n).yield_dataframe_as("result")

    rows = list(
        dag.run(
            execution_engine,
            conf=execution_engine_conf,
        )["result"].as_dict_iterable()
    )
    return [
        TrialReport.from_jsondict(json.loads(r[TUNE_REPORT]))
        for r in sorted(rows, key=lambda r: r[TUNE_REPORT_METRIC])
    ]


def suggest_sk_models_by_cv(
    space: Space,
    train_df: Any,
    scoring: str,
    cv: int = 5,
    temp_path: str = "",
    feature_prefix: str = "",
    label_col: str = "label",
    save_model: bool = False,
    partition_keys: Optional[List[str]] = None,
    top_n: int = 1,
    local_optimizer: Optional[NonIterativeObjectiveLocalOptimizer] = None,
    monitor: Any = None,
    stopper: Any = None,
    stop_check_interval: Any = None,
    distributed: Optional[bool] = None,
    execution_engine: Any = None,
    execution_engine_conf: Any = None,
) -> List[TrialReport]:
    dag = FugueWorkflow()
    dataset = TUNE_OBJECT_FACTORY.make_dataset(
        dag,
        space,
        df=train_df,
        partition_keys=partition_keys,
        temp_path=temp_path,
    )
    objective = SKCVObjective(
        scoring=scoring,
        cv=cv,
        feature_prefix=feature_prefix,
        label_col=label_col,
        checkpoint_path=temp_path if save_model else None,
    )
    study = optimize_noniterative(
        objective=objective,
        dataset=dataset,
        optimizer=local_optimizer,
        distributed=distributed,
        monitor=monitor,
        stopper=stopper,
        stop_check_interval=stop_check_interval,
    )
    study.result(top_n).yield_dataframe_as("result")

    rows = list(
        dag.run(
            execution_engine,
            conf=execution_engine_conf,
        )["result"].as_dict_iterable()
    )
    return [
        TrialReport.from_jsondict(json.loads(r[TUNE_REPORT]))
        for r in sorted(rows, key=lambda r: r[TUNE_REPORT_METRIC])
    ]



from threading import RLock
from typing import Any, Callable, Dict, List, Optional, Tuple

import optuna
from optuna.study import Study
from tune import (
    NonIterativeObjectiveFunc,
    NonIterativeObjectiveLocalOptimizer,
    Rand,
    RandInt,
    StochasticExpression,
    Trial,
    TrialReport,
)
from tune._utils.math import _IGNORABLE_ERROR, uniform_to_discrete, uniform_to_integers
from tune.concepts.space.parameters import Choice


class OptunaLocalOptimizer(NonIterativeObjectiveLocalOptimizer):
    def __init__(
        self, max_iter: int, create_study: Optional[Callable[[], Study]] = None
    ):
        self._max_iter = max_iter
        self._create_study = create_study or optuna.create_study

    def run(self, func: NonIterativeObjectiveFunc, trial: Trial) -> TrialReport:
        static_params, stochastic_params = _split(trial.params)
        if len(stochastic_params) == 0:
            return func.run(trial)
        lock = RLock()
        best_report: List[TrialReport] = []

        def obj(otrial: optuna.trial.Trial) -> float:
            params = _convert(otrial, stochastic_params)
            params.update(static_params)
            report = func.run(trial.with_params(params))
            with lock:
                if len(best_report) == 0:
                    best_report.append(report)
                elif report.sort_metric < best_report[0].sort_metric:
                    best_report[0] = report
            return report.sort_metric

        study = self._create_study()
        study.optimize(obj, n_trials=self._max_iter)
        assert 1 == len(best_report)
        return best_report[0]


def _split(
    kwargs: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, StochasticExpression]]:
    static_params: Dict[str, Any] = {}
    stochastic_params: Dict[str, Any] = {}
    for k, v in kwargs.items():
        if isinstance(v, StochasticExpression):
            stochastic_params[k] = v
        else:
            static_params[k] = v
    return static_params, stochastic_params


def _convert(
    trial: optuna.trial.Trial, params: Dict[str, StochasticExpression]
) -> Dict[str, Any]:
    result: Dict[str, Any] = {}
    for k, v in params.items():
        if isinstance(v, RandInt):
            if v.log and v.q is not None:
                value = trial.suggest_float(name=k, low=0, high=1.0)
                result[k] = uniform_to_integers(
                    value,
                    low=v.low,
                    high=v.high,
                    q=v.q,  # type: ignore
                    log=True,
                    include_high=v.include_high,
                )
            else:
                _high: Any = v.high if v.include_high else v.high - 1
                result[k] = trial.suggest_int(
                    name=k, low=v.low, high=_high, step=v.q, log=v.log
                )
        elif isinstance(v, Rand):
            if v.log and v.q is not None:
                value = trial.suggest_float(name=k, low=0, high=1.0)
                result[k] = uniform_to_discrete(
                    value,
                    low=v.low,
                    high=v.high,
                    q=v.q,
                    log=True,
                    include_high=v.include_high,
                )
            else:
                _high = v.high
                if v.q is not None and not v.include_high:
                    _high -= _IGNORABLE_ERROR
                result[k] = trial.suggest_float(
                    name=k, low=v.low, high=_high, step=v.q, log=v.log
                )
        elif isinstance(v, Choice):
            result[k] = trial.suggest_categorical(name=k, choices=v.values)
        else:  # pragma: no cover
            raise NotImplementedError
    return result
